/*

Simulating the melting color glass condensate (glasma) -- 3d

by Paul Romatschke

this code was used and is documented in

P.~Romatschke and R.~Venugopalan, 
* arxiv:hep-ph/0510121 (Phys.Rev.Lett.96:062302,2006)
* arxiv:hep-ph/0510292 (Eur.Phys.J.A29:71-75,2006)
* arxiv:hep-ph/0605045 (Phys.Rev.D74:045011,2006).

If you use this code or some part of it, be sure to reference these
articles.

Copyright of this code is granted provided you keep this disclaimer.

*/
#include <iostream>
#include <iomanip>
#include <fstream>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <DoublePrn.cpp>
#include <su2mat.h>

using namespace std;

#define SQRT3		sqrt(3.0)
#define SQRT2		sqrt(2.0)


// these global vars are initialized from parameters file
// defaults set here are overridden by that file
int    NUM=20,NUMT=20, STEPS=40000,UPDATE=100,SNAPUPDATE=1000;
int    USEPOLYHEDRA=1,NW=8,NPHI=5,NZ=10,NTRAJ=10;
int    MIMICK2D=1,RANDOM=1,RANGEN=1;
int    MIDPOINT;
double A=0.05,EPS=0.001,XI=10,MU2=10,MAXEVIOLATION=1,TSIG=1,TINIT=0.002;


// these hold the current values; note w[v] = a[v] wo[v] + b[v] wz[v]  
//SU2matrix *phix,*phiy,
SU2matrix ***pix,***piy,***ez;
SU2LieMatrix ***u,***ux,***uy;

// these hold the updated values
//SU2matrix *Phix,*Phiy,
SU2matrix ***Pix,***Piy,***Ez;
SU2LieMatrix ***U,***Ux,***Uy;

//for initialization we need color sources and potentials
SU2matrix **Lambda1,**Lambda2,**Lambda3;

double *rho1a,*rho1b,*rho1c;
double *rho2a,*rho2b,*rho2c;
double *rho3a,*rho3b,*rho3c;
double *f1,*f2;


double *FTarray,*FT1darray;

// j holds current at t; oj holds current at t-eps
SU2matrix ****j,****oj;

// vertex array - x=0,y=1,z=2 - and coefficient arrays - a and b
double **v,*a,*b;

// these are used during the computation of the abelianization correlation
// length; uxiphi's hold the values of phi's parallel transported to the
// right by x numbers of steps; xiA and xiPhi hold the resulting correlation lengths
// likewise for uxij's and associated xi's
SU2matrix ***uxiphix,***uxiphiy,***uxijx,***uxijy;
double xiA,xiPhi,xiAJ,xiJ;

//Initial Gauss law violation
SU2matrix InitialGLv;

//overall time
double otime = 0;

// InitialEnergy of the system
double InitialEnergy = 0;

// Holds the percentage difference from the initial energy
double energyDiff = 0;

// stores the running integrated value of the HL contribution to the energy
double HLenergy = 0;
double eloss=0;

// output files
fstream phirms_out;
fstream ED_out;
fstream jrms_out;
fstream cbar_out;
fstream energy_out;
fstream gausslaw_out;
fstream clengths_out;


// initialize global arrays
void allocateMemory() {

    	cout << "==> Allocating memory\n";

	pix = new SU2matrix**[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) pix[sx] = new SU2matrix*[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) pix[sx][sy] = new SU2matrix[NUM+2];

	piy = new SU2matrix**[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) piy[sx] = new SU2matrix*[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) piy[sx][sy] = new SU2matrix[NUM+2];

	ez = new SU2matrix**[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) ez[sx] = new SU2matrix*[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) ez[sx][sy] = new SU2matrix[NUM+2];

	u = new SU2LieMatrix**[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) u[sx] = new SU2LieMatrix*[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) u[sx][sy] = new SU2LieMatrix[NUM+2];

	ux = new SU2LieMatrix**[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) ux[sx] = new SU2LieMatrix*[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) ux[sx][sy] = new SU2LieMatrix[NUM+2];

	uy = new SU2LieMatrix**[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) uy[sx] = new SU2LieMatrix*[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) uy[sx][sy] = new SU2LieMatrix[NUM+2];

	


	Pix = new SU2matrix**[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) Pix[sx] = new SU2matrix*[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) Pix[sx][sy] = new SU2matrix[NUM+2];

	Piy = new SU2matrix**[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) Piy[sx] = new SU2matrix*[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) Piy[sx][sy] = new SU2matrix[NUM+2];

	Ez = new SU2matrix**[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) Ez[sx] = new SU2matrix*[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) Ez[sx][sy] = new SU2matrix[NUM+2];

	U = new SU2LieMatrix**[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) U[sx] = new SU2LieMatrix*[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) U[sx][sy] = new SU2LieMatrix[NUM+2];

	Ux = new SU2LieMatrix**[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) Ux[sx] = new SU2LieMatrix*[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) Ux[sx][sy] = new SU2LieMatrix[NUM+2];

	Uy = new SU2LieMatrix**[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) Uy[sx] = new SU2LieMatrix*[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) Uy[sx][sy] = new SU2LieMatrix[NUM+2];


	
	rho1a = new double[(NUMT)*(NUMT)];
	rho1b = new double[(NUMT)*(NUMT)];
	rho1c = new double[(NUMT)*(NUMT)];

	f1    = new double[NUM+2];
	f2    = new double[NUM+2];

	rho2a = new double[(NUMT)*(NUMT)];
	rho2b = new double[(NUMT)*(NUMT)];
	rho2c = new double[(NUMT)*(NUMT)];

	rho3a = new double[(NUMT)*(NUMT)];
	rho3b = new double[(NUMT)*(NUMT)];
	rho3c = new double[(NUMT)*(NUMT)];

	FTarray = new double[(NUMT)*(NUMT)*2];
	FT1darray = new double[NUM*2];

	Lambda1 = new SU2matrix*[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) Lambda1[sx] = new SU2matrix[NUMT+2];

	Lambda2 = new SU2matrix*[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) Lambda2[sx] = new SU2matrix[NUMT+2];

	Lambda3 = new SU2matrix*[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) Lambda3[sx] = new SU2matrix[NUMT+2];
	
	return;
}

// enforces boundary conditions by explicitly copying values into the padding elements
void enforceBCs()
{
  SU2matrix tempf,templ;
//fill in z=0,NUM+1 values for main x,y
  for (int sx=1;sx<=NUMT;sx++) 
    for (int sy=1;sy<=NUMT;sy++) { 

	pix[sx][sy][0] = pix[sx][sy][NUM]; 	pix[sx][sy][NUM+1] = pix[sx][sy][1];
	piy[sx][sy][0] = piy[sx][sy][NUM]; 	piy[sx][sy][NUM+1] = piy[sx][sy][1];
	ez[sx][sy][0] = ez[sx][sy][NUM]; 	ez[sx][sy][NUM+1] = ez[sx][sy][1];
	u[sx][sy][0] = u[sx][sy][NUM]; 		u[sx][sy][NUM+1] = u[sx][sy][1];
	ux[sx][sy][0] = ux[sx][sy][NUM]; 	ux[sx][sy][NUM+1] = ux[sx][sy][1];
	uy[sx][sy][0] = uy[sx][sy][NUM]; 	uy[sx][sy][NUM+1] = uy[sx][sy][1];      
    }
  
  //now fill in y=0,NUMT+1 values for main x and all z
  for (int sx=1;sx<=NUMT;sx++) 
    for (int sz=0;sz<=NUM+1;sz++) { 
      pix[sx][0][sz] = pix[sx][NUMT][sz]; 	pix[sx][NUMT+1][sz] = pix[sx][1][sz];
	piy[sx][0][sz] = piy[sx][NUMT][sz]; 	piy[sx][NUMT+1][sz] = piy[sx][1][sz];
	ez[sx][0][sz] = ez[sx][NUMT][sz]; 	ez[sx][NUMT+1][sz] = ez[sx][1][sz];
	u[sx][0][sz] = u[sx][NUMT][sz]; 	u[sx][NUMT+1][sz] = u[sx][1][sz];
	ux[sx][0][sz] = ux[sx][NUMT][sz]; 	ux[sx][NUMT+1][sz] = ux[sx][1][sz];
	uy[sx][0][sz] = uy[sx][NUMT][sz]; 	uy[sx][NUMT+1][sz] = uy[sx][1][sz];
    }
  
  //finally fill in x=0,NUMT+1 values for general y,z

  for (int sy=0;sy<=NUMT+1;sy++) 
    for (int sz=0;sz<=NUM+1;sz++) { 
      pix[0][sy][sz] = pix[NUMT][sy][sz]; 	pix[NUMT+1][sy][sz] = pix[1][sy][sz];
      piy[0][sy][sz] = piy[NUMT][sy][sz]; 	piy[NUMT+1][sy][sz] = piy[1][sy][sz];
      ez[0][sy][sz] = ez[NUMT][sy][sz]; 	ez[NUMT+1][sy][sz] = ez[1][sy][sz];
      u[0][sy][sz] = u[NUMT][sy][sz]; 		u[NUMT+1][sy][sz] = u[1][sy][sz];
      ux[0][sy][sz] = ux[NUMT][sy][sz]; 	ux[NUMT+1][sy][sz] = ux[1][sy][sz];
      uy[0][sy][sz] = uy[NUMT][sy][sz]; 	uy[NUMT+1][sy][sz] = uy[1][sy][sz];
    }

}

/*
 * copies Pi,Phi,Ez,U,W to pi,phi,ez,u,w and then calls enforceBCs()
 *
*/
void copyDown() {
for (int sx=1;sx<=NUMT;sx++) 
    for (int sy=1;sy<=NUMT;sy++)
      for (int sz=1; sz<=NUM;sz++) { 
	pix[sx][sy][sz] = Pix[sx][sy][sz];
	piy[sx][sy][sz] = Piy[sx][sy][sz];
	ez[sx][sy][sz] = Ez[sx][sy][sz];
	u[sx][sy][sz] = U[sx][sy][sz];
	ux[sx][sy][sz] = Ux[sx][sy][sz];
	uy[sx][sy][sz] = Uy[sx][sy][sz];
      }
 enforceBCs();
}



//fixes Coulomb gauge in 2d

void gaugefix2d(int maxiter)
{
  

  //assumes that ux[],... hold matching values

  //FT
  extern void FT(double* pArr, int nx, int ny,int sign);

  int nx=NUMT;
  int ny=NUMT;
  int l=0;

  SU2LieMatrix** gux = new SU2LieMatrix*[NUMT+2]; 
  for (int sx=0;sx<NUMT+2;sx++) gux[sx] = new SU2LieMatrix[NUMT+2];



  for (int sx=1;sx<=NUMT;sx++) 
    for (int sy=1;sy<=NUMT;sy++)
      {
	gux[sx][sy].a=1;
	gux[sx][sy].b=0;
	gux[sx][sy].c=0;
	gux[sx][sy].d=0;
	Lambda1[sx][sy]=Adjoint(gux[sx][sy]);
      }

  for (int sx=1;sx<=NUMT;sx++) 
	{gux[sx][0] = gux[sx][NUMT]; 	gux[sx][NUMT+1]= gux[sx][1];}
      for (int sy=0;sy<=NUMT+1;sy++) 
	{gux[0][sy] = gux[NUMT][sy]; 	gux[NUMT+1][sy]= gux[1][sy];}

  double conv=1;
  for (int gfiter=0;gfiter<maxiter;gfiter++)
    {

      if (conv/NUMT/NUMT/NUM<1.e-15)
	break;

      conv=0;

      //generate divergences
      for (int sx=1;sx<=NUMT;sx++) 
	for (int sy=1;sy<=NUMT;sy++)
	  {
	
	    Lambda1[sx][sy]=Adjoint(Hermitian(gux[sx][sy])*ux[sx][sy][1]*gux[sx+1][sy]);
	    Lambda1[sx][sy]-=Adjoint(Hermitian(gux[sx-1][sy])*ux[sx-1][sy][1]*gux[sx][sy]);
	    Lambda1[sx][sy]+=Adjoint(Hermitian(gux[sx][sy])*uy[sx][sy][1]*gux[sx][sy+1]);
	    Lambda1[sx][sy]-=Adjoint(Hermitian(gux[sx][sy-1])*uy[sx][sy-1][1]*gux[sx][sy]);


	    //Lambda1[sx][sy]=Adjoint(ux[sx][sy][1]-ux[sx-1][sy][1]+uy[sx][sy][1]-uy[sx][sy-1][1]);
	    conv+=tr2(Lambda1[sx][sy]);

	    FTarray[2*((sx-1)*(NUMT)+sy-1)]=Lambda1[sx][sy].a;
	    FTarray[2*((sx-1)*(NUMT)+sy-1)+1]=0;
	  }

      //cout << "Converging:  " << conv/NUMT/NUMT << endl;

      //    conv=0;
      //      for (int sx=1;sx<=NUMT;sx++) 
      //for (int sy=1;sy<=NUMT;sy++)
      //{
      //  conv+=Lambda1[sx][sy].a;
      //}
      //cout << "Constant part" << conv/NUMT/NUMT << endl;

      //Lambda1[2][2].print();

//solve Poisson's equation: (L1a)

      FT(FTarray,NUMT,NUMT,1);
      
      FTarray[0]=0; // remove constant part
      FTarray[1]=0;
      conv=0;
      for (int j=0;j<nx;j++) //solve for X
	for (int k=0;k<ny;k++)
	  {
	    l=k+j*nx;
	    l=(l<<1);
	    if (l!=0)
	      {
		FTarray[l]=FTarray[l]/2/(cos(6.28318530717958647692529*j/nx)+cos(6.28318530717958647692529*k/ny)-2);
		l++;
		FTarray[l]=FTarray[l]/2/(cos(6.28318530717958647692529*j/nx)+cos(6.28318530717958647692529*k/ny)-2);
	      }
	    }
      
      FT(FTarray,NUMT,NUMT,-1); //transform back
      for (int sx=0;sx<NUMT;sx++) //paste to Lambda's
	for (int sy=0;sy<NUMT;sy++)
	  {
	    Lambda1[sx+1][sy+1].a=FTarray[2*(sx*(NUMT)+sy)];
	    //conv+=FTarray[2*(sx*(NUMT)+sy)];
	  }
      //cout << "another test: " << conv << endl;
      //Lambda1[2][2].print();
      
      //next (L1b)

      
      for (int sx=1;sx<=NUMT;sx++) 
	for (int sy=1;sy<=NUMT;sy++)
	  {
	    FTarray[2*((sx-1)*(NUMT)+sy-1)]=Lambda1[sx][sy].b;
	    FTarray[2*((sx-1)*(NUMT)+sy-1)+1]=0;
	  }
      FT(FTarray,NUMT,NUMT,1);
      FTarray[0]=0; // remove constant part
      FTarray[1]=0;
      for (int j=0;j<nx;j++) //solve for X
	for (int k=0;k<ny;k++)
	  {
	    l=k+j*nx;
	    l=(l<<1);
	    if (l!=0)
	      {
		FTarray[l]=FTarray[l]/2/(cos(6.28318530717958647692529*j/nx)+cos(6.28318530717958647692529*k/ny)-2);
		l++;
		FTarray[l]=FTarray[l]/2/(cos(6.28318530717958647692529*j/nx)+cos(6.28318530717958647692529*k/ny)-2);
	      }
	  }
      FT(FTarray,NUMT,NUMT,-1); //transform back
      for (int sx=0;sx<NUMT;sx++) //paste to Lambda's
	for (int sy=0;sy<NUMT;sy++)
	    Lambda1[sx+1][sy+1].b=FTarray[2*(sx*(NUMT)+sy)];

      //next(L1c)
      
      for (int sx=1;sx<=NUMT;sx++) 
	for (int sy=1;sy<=NUMT;sy++)
	  {
	    FTarray[2*((sx-1)*(NUMT)+sy-1)]=Lambda1[sx][sy].c;
	    FTarray[2*((sx-1)*(NUMT)+sy-1)+1]=0;
	  }
      FT(FTarray,NUMT,NUMT,1);
      FTarray[0]=0; // remove constant part
      FTarray[1]=0;
      for (int j=0;j<nx;j++) //solve for X
	for (int k=0;k<ny;k++)
	  {
	    l=k+j*nx;
	    l=(l<<1);
	    if (l!=0)
	      {
		FTarray[l]=FTarray[l]/2/(cos(6.28318530717958647692529*j/nx)+cos(6.28318530717958647692529*k/ny)-2);
		l++;
		FTarray[l]=FTarray[l]/2/(cos(6.28318530717958647692529*j/nx)+cos(6.28318530717958647692529*k/ny)-2);
	      }
	  }
      FT(FTarray,NUMT,NUMT,-1); //transform back
      for (int sx=0;sx<NUMT;sx++) //paste to Lambda's
	for (int sy=0;sy<NUMT;sy++)
	    Lambda1[sx+1][sy+1].c=FTarray[2*(sx*(NUMT)+sy)];

      
      //(Lambda1[3][2]+Lambda1[1][2]+Lambda1[2][3]+Lambda1[2][1]-4*Lambda1[2][2]).print();
 

      //prepare gauge transform

      conv=0;
      for (int sx=1;sx<=NUMT;sx++) 
	for (int sy=1;sy<=NUMT;sy++)
	  {
	    gux[sx][sy]=gux[sx][sy]*mexpi(Lambda1[sx][sy]*(-1.0));  // V1
	    gux[sx][sy].normalize();
	    conv+=tr2(Lambda1[sx][sy]);
	  }
      
      //cout << " X: " << conv/NUMT/NUMT << endl;   

      for (int sx=1;sx<=NUMT;sx++) 
	{gux[sx][0] = gux[sx][NUMT]; 	gux[sx][NUMT+1]= gux[sx][1];}
      for (int sy=0;sy<=NUMT+1;sy++) 
	{gux[0][sy] = gux[NUMT][sy]; 	gux[NUMT+1][sy]= gux[1][sy];}
      

      //conv=0;

      //          cout << "newtry" << endl;

      //Adjoint(ux[3][2][1]+uy[2][3][1]-uy[2][2][1]-ux[2][2][1]).print();
      //Adjoint(ux[2][2][1]).print();
      //(Lambda1[3][2]-Lambda1[2][2]).print();
      //(Adjoint(ux[2][2][1])-Lambda1[3][2]/10+Lambda1[2][2]/10).print();
      //Adjoint(Hermitian(gux[2][2][3])*ux[2][2][1]*gux[3][2][3]).print();
      //Adjoint(Hermitian(gux[2][2][3])*ux[2][2][1]*gux[3][2][3]-ux[2][2][1]).print();
      //Adjoint(gux[3][2][3]-gux[2][2][3]).print();
      //(Adjoint(gux[3][2][3]+gux[1][2][3]+gux[2][3][3]+gux[2][1][3])-Adjoint(gux[2][2][3])*4).print();

      

      //cout << " X2: " << conv/NUMT/NUMT << endl;   


      //(Adjoint(gux[3][2][3]+gux[1][2][3]+gux[2][3][3]+gux[2][1][3])-Adjoint(gux[2][2][3])*4).print();

      //enforceBCs();

    } // end of main loop

  //cout << "Hello?" << endl;
  
  //do gauge Transform

  for (int sx=1;sx<=NUMT;sx++) 
    for (int sy=1;sy<=NUMT;sy++)
      for (int sz=1;sz<=NUM;sz++)
	{
	  ux[sx][sy][sz]=Hermitian(gux[sx][sy])*ux[sx][sy][sz]*gux[sx+1][sy];  //these are correct ones
	  uy[sx][sy][sz]=Hermitian(gux[sx][sy])*uy[sx][sy][sz]*gux[sx][sy+1];
	  u[sx][sy][sz]=Hermitian(gux[sx][sy])*u[sx][sy][sz]*gux[sx][sy];
	  pix[sx][sy][sz]=transportLeft(pix[sx][sy][sz],gux[sx][sy]);
	  piy[sx][sy][sz]=transportLeft(piy[sx][sy][sz],gux[sx][sy]);
	  ez[sx][sy][sz]=transportLeft(ez[sx][sy][sz],gux[sx][sy]);
	  
	  ux[sx][sy][sz]/=ux[sx][sy][sz].getNorm();  //reunitarize
	  uy[sx][sy][sz]/=uy[sx][sy][sz].getNorm();
	  u[sx][sy][sz]/=u[sx][sy][sz].getNorm();
	}
	    
  enforceBCs();

  for (int sx=0;sx<NUMT+2;sx++) 
    delete [] gux[sx];

}


//fixes Coulomb gauge in 3d

void gaugefix3d(int maxiter,double time)
{
  
  //FT
  extern void FT3d(double* pArr, int nx, int ny, int nz, int sign);

  int nx=NUMT;
  int ny=NUMT;
  int nz=NUM;
  int l=0;
  double* FT3darray = new double[NUMT*NUMT*NUM*2];
  
  SU2LieMatrix ***gux;
  gux = new SU2LieMatrix**[NUMT+2]; 
  for (int sx=0;sx<NUMT+2;sx++) gux[sx] = new SU2LieMatrix*[NUMT+2];
  for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) gux[sx][sy] = new SU2LieMatrix[NUM+2];

  
  for (int sx=1;sx<=NUMT;sx++) 
    for (int sy=1;sy<=NUMT;sy++)
      for (int sz=1;sz<=NUM;sz++)
      {
	gux[sx][sy][sz].a=1;
	gux[sx][sy][sz].b=0;
	gux[sx][sy][sz].c=0;
	gux[sx][sy][sz].d=0;
      }


  //BCs
  for (int sx=1;sx<=NUMT;sx++) 
    for (int sy=1;sy<=NUMT;sy++) { 
	gux[sx][sy][0] = gux[sx][sy][NUM]; 	gux[sx][sy][NUM+1] = gux[sx][sy][1];
    }
  for (int sx=1;sx<=NUMT;sx++) 
    for (int sz=0;sz<=NUM+1;sz++) { 
	gux[sx][0][sz] = gux[sx][NUMT][sz]; 	gux[sx][NUMT+1][sz] = gux[sx][1][sz];
    }
  for (int sy=0;sy<=NUMT+1;sy++) 
    for (int sz=0;sz<=NUM+1;sz++) { 
      gux[0][sy][sz] = gux[NUMT][sy][sz]; 	gux[NUMT+1][sy][sz] = gux[1][sy][sz];
    }

  


  double conv=1;
  for (int gfiter=0;gfiter<maxiter;gfiter++)
    { 
      if (conv/NUMT/NUMT/NUM<1.e-15)
	break;

      conv=0;

      //generate divergences
      for (int sx=1;sx<=NUMT;sx++) 
	for (int sy=1;sy<=NUMT;sy++)
	  for (int sz=1;sz<=NUM;sz++)
	  {
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]=Adjoint(u[sx][sy][sz]).a/A/time/time;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]-=Adjoint(u[sx][sy][sz-1]).a/A/time/time;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]+=Adjoint(ux[sx][sy][sz]).a;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]-=Adjoint(ux[sx-1][sy][sz]).a;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]+=Adjoint(uy[sx][sy][sz]).a;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]-=Adjoint(uy[sx][sy-1][sz]).a;
	    

	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))+1]=0.0;

	    conv+=FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]*FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))];
	    
	  }
      
      //solve Poisson's equation: (L1a)

      FT3d(FT3darray,NUMT,NUMT,NUM,1);

      
      FT3darray[0]=0; // remove constant part
      FT3darray[1]=0;
      //conv=0;
      for (int j=0;j<nx;j++) //solve for X
	for (int k=0;k<ny;k++)
	  for (int m=0;m<nz;m++)
	  {
	    l=m+k*nz+j*ny*nz;
	    l=(l<<1);
	    if (l!=0)
	      {
		FT3darray[l]=FT3darray[l]/2/(cos(2*M_PI*j/nx)+cos(2*M_PI*k/ny)-2+(cos(2*M_PI*m/nz)-1)/time/time/A/A);
		l++;
		FT3darray[l]=FT3darray[l]/2/(cos(2*M_PI*j/nx)+cos(2*M_PI*k/ny)-2+(cos(2*M_PI*m/nz)-1)/time/time/A/A);
	      }
	    }
      
      FT3d(FT3darray,NUMT,NUMT,NUM,-1); //transform back


      //assume GT has been done already and gux is empty
      for (int sx=0;sx<NUMT;sx++) 
	for (int sy=0;sy<NUMT;sy++)
	  for (int sz=0;sz<NUM;sz++)
	    gux[sx+1][sy+1][sz+1].b=FT3darray[2*((sx)*NUMT*NUM+(sy)*NUM+(sz))];

      for (int sx=1;sx<=NUMT;sx++) 
	for (int sy=1;sy<=NUMT;sy++)
	  for (int sz=1;sz<=NUM;sz++)
	  {
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]=Adjoint(ux[sx][sy][sz]).b;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]-=Adjoint(ux[sx-1][sy][sz]).b;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]+=Adjoint(uy[sx][sy][sz]).b;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]-=Adjoint(uy[sx][sy-1][sz]).b;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]+=Adjoint(u[sx][sy][sz]).b/A/time/time;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]-=Adjoint(u[sx][sy][sz-1]).b/A/time/time;

	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))+1]=0.0;

	    conv+=FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]*FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))];
	    
	  }

      FT3d(FT3darray,NUMT,NUMT,NUM,1);

      FT3darray[0]=0; // remove constant part
      FT3darray[1]=0;
      //conv=0;
      for (int j=0;j<nx;j++) //solve for X
	for (int k=0;k<ny;k++)
	  for (int m=0;m<nz;m++)
	  {
	    l=m+k*nz+j*ny*nz;
	    l=(l<<1);
	    if (l!=0)
	      {
		FT3darray[l]=FT3darray[l]/2/(cos(2*M_PI*j/nx)+cos(2*M_PI*k/ny)-2+(cos(2*M_PI*m/nz)-1)/time/time/A/A);
		l++;
		FT3darray[l]=FT3darray[l]/2/(cos(2*M_PI*j/nx)+cos(2*M_PI*k/ny)-2+(cos(2*M_PI*m/nz)-1)/time/time/A/A);
	      }
	    }
      
      FT3d(FT3darray,NUMT,NUMT,NUM,-1); //transform back


      //assume GT has been done already and gux is empty
      for (int sx=0;sx<NUMT;sx++) 
	for (int sy=0;sy<NUMT;sy++)
	  for (int sz=0;sz<NUM;sz++)
	    gux[sx+1][sy+1][sz+1].c=FT3darray[2*((sx)*NUMT*NUM+(sy)*NUM+(sz))];



      for (int sx=1;sx<=NUMT;sx++) 
	for (int sy=1;sy<=NUMT;sy++)
	  for (int sz=1;sz<=NUM;sz++)
	  {
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]=Adjoint(ux[sx][sy][sz]).c;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]-=Adjoint(ux[sx-1][sy][sz]).c;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]+=Adjoint(uy[sx][sy][sz]).c;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]-=Adjoint(uy[sx][sy-1][sz]).c;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]+=Adjoint(u[sx][sy][sz]).c/A/time/time;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]-=Adjoint(u[sx][sy][sz-1]).c/A/time/time;

	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))+1]=0.0;

	    conv+=FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]*FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))];
	  }

      //solve Poisson's equation: (L1c)

      FT3d(FT3darray,NUMT,NUMT,NUM,1);

      
      FT3darray[0]=0; // remove constant part
      FT3darray[1]=0;
      //conv=0;
      for (int j=0;j<nx;j++) //solve for X
	for (int k=0;k<ny;k++)
	  for (int m=0;m<nz;m++)
	  {
	    l=m+k*nz+j*ny*nz;
	    l=(l<<1);
	    if (l!=0)
	      {
		FT3darray[l]=FT3darray[l]/2/(cos(2*M_PI*j/nx)+cos(2*M_PI*k/ny)-2+(cos(2*M_PI*m/nz)-1)/time/time/A/A);
		l++;
		FT3darray[l]=FT3darray[l]/2/(cos(2*M_PI*j/nx)+cos(2*M_PI*k/ny)-2+(cos(2*M_PI*m/nz)-1)/time/time/A/A);
	      }
	    }
      
      FT3d(FT3darray,NUMT,NUMT,NUM,-1); //transform back

      //assume GT has been done already and gux is empty
      for (int sx=0;sx<NUMT;sx++) 
	for (int sy=0;sy<NUMT;sy++)
	  for (int sz=0;sz<NUM;sz++)
	    gux[sx+1][sy+1][sz+1].d=FT3darray[2*((sx)*NUMT*NUM+(sy)*NUM+(sz))];
      
      //printf("just for debugging: %e\n",conv/NUMT/NUMT/NUM);

      //prepare gauge transform

      SU2matrix temp;
      for (int sx=1;sx<=NUMT;sx++) 
	for (int sy=1;sy<=NUMT;sy++)
	  for (int sz=1;sz<=NUM;sz++)
	  {
	    temp.a=gux[sx][sy][sz].b;
	    temp.b=gux[sx][sy][sz].c;
	    temp.c=gux[sx][sy][sz].d;
	    gux[sx][sy][sz]=mexpi(temp*(-1.0));  // V1
	    gux[sx][sy][sz].normalize();
	  }

      

      //BCs
      for (int sx=1;sx<=NUMT;sx++) 
	for (int sy=1;sy<=NUMT;sy++) { 
	  gux[sx][sy][0] = gux[sx][sy][NUM]; 	gux[sx][sy][NUM+1] = gux[sx][sy][1];
	}
      for (int sx=1;sx<=NUMT;sx++) 
	for (int sz=0;sz<=NUM+1;sz++) { 
	  gux[sx][0][sz] = gux[sx][NUMT][sz]; 	gux[sx][NUMT+1][sz] = gux[sx][1][sz];
	}
      for (int sy=0;sy<=NUMT+1;sy++) 
	for (int sz=0;sz<=NUM+1;sz++) { 
	  gux[0][sy][sz] = gux[NUMT][sy][sz]; 	gux[NUMT+1][sy][sz] = gux[1][sy][sz];
	}

      //do gauge Transform

      for (int sx=1;sx<=NUMT;sx++) 
	for (int sy=1;sy<=NUMT;sy++)
	  for (int sz=1;sz<=NUM;sz++)
	    {
	      ux[sx][sy][sz]=Hermitian(gux[sx][sy][sz])*ux[sx][sy][sz]*gux[sx+1][sy][sz];
	      uy[sx][sy][sz]=Hermitian(gux[sx][sy][sz])*uy[sx][sy][sz]*gux[sx][sy+1][sz];
	      u[sx][sy][sz]=Hermitian(gux[sx][sy][sz])*u[sx][sy][sz]*gux[sx][sy][sz+1];
	      pix[sx][sy][sz]=transportLeft(pix[sx][sy][sz],gux[sx][sy][sz]);
	      piy[sx][sy][sz]=transportLeft(piy[sx][sy][sz],gux[sx][sy][sz]);
	      ez[sx][sy][sz]=transportLeft(ez[sx][sy][sz],gux[sx][sy][sz]);
	      
	      ux[sx][sy][sz]/=ux[sx][sy][sz].getNorm();  //reunitarize
	      uy[sx][sy][sz]/=uy[sx][sy][sz].getNorm();
	      u[sx][sy][sz]/=u[sx][sy][sz].getNorm();
	    }
      
      enforceBCs();
      
    } // end of main loop

  //just for debugging:
  //printf("debug: %e\n",conv/NUMT/NUMT/NUM);

  if (conv/NUMT/NUMT/NUM>1.e-15)
    {
      cout << "GF warning: did not converge at "<< maxiter << "iterations, conv = ";
      cout <<conv/NUMT/NUMT/NUM << endl;
    }
  delete [] FT3darray;

  for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) 
    delete [] gux[sx][sy];

}


void doMatching2d()
{
  double resa, resb, resc, resd;
  resa=0;
  resb=0;
  resc=0;
  resd=0;
	    
  SU2LieMatrix u1x,u2x,u1y,u2y,bu1x,bu2x,bu1y,bu2y;
  
  SU2matrix temp1,temp2;

  SU2LieMatrix utemp;
  
  double checknorm2=0;
  
  double checknorm=0;

  for (int sx=1;sx<=NUMT;sx++) 
    for (int sy=1;sy<=NUMT;sy++)
      {
	
	u1x=Hermitian(mexpi(Lambda1[sx][sy]))*mexpi(Lambda1[sx+1][sy]);	      
	u2x=Hermitian(mexpi(Lambda2[sx][sy]))*mexpi(Lambda2[sx+1][sy]);
	u1y=Hermitian(mexpi(Lambda1[sx][sy]))*mexpi(Lambda1[sx][sy+1]);
	u2y=Hermitian(mexpi(Lambda2[sx][sy]))*mexpi(Lambda2[sx][sy+1]);
	u1x.normalize();
	u2x.normalize();
	u1y.normalize();
	u2y.normalize();	  

		  
	for (int sz=1; sz<=NUM;sz++)
	  {	    
	    u[sx][sy][sz].a = 1;
	    utemp=Hermitian(u1x)+Hermitian(u2x);
	    ux[sx][sy][sz]=(u1x+u2x)*Hermitian(utemp/(utemp.getNorm()))/(utemp.getNorm());
		      //checknorm+=ux[sx][sy][sz].getNorm()-1;
	    utemp=Hermitian(u1y)+Hermitian(u2y);
	    uy[sx][sy][sz]=(u1y+u2y)*Hermitian(utemp/utemp.getNorm())/utemp.getNorm();
		      //checknorm+=uy[sx][sy][sz].getNorm()-1;
	    ux[sx][sy][sz].normalize();
	    uy[sx][sy][sz].normalize();

	  }
      }

	
  enforceBCs(); //make sx-1 and so on possible
	    
  resc=0;
  for (int sx=1;sx<=NUMT;sx++) 
    for (int sy=1;sy<=NUMT;sy++)
      {
		  
	u1x=Hermitian(mexpi(Lambda1[sx][sy]))*mexpi(Lambda1[sx+1][sy]);
	u2x=Hermitian(mexpi(Lambda2[sx][sy]))*mexpi(Lambda2[sx+1][sy]);
	u1y=Hermitian(mexpi(Lambda1[sx][sy]))*mexpi(Lambda1[sx][sy+1]);
	u2y=Hermitian(mexpi(Lambda2[sx][sy]))*mexpi(Lambda2[sx][sy+1]);
	u1x/=u1x.getNorm();
	u1y/=u1y.getNorm();
	u2x/=u2x.getNorm();
	u2y/=u2y.getNorm();
		  
		  
	bu1x=Hermitian(mexpi(Lambda1[sx-1][sy]))*mexpi(Lambda1[sx][sy]);
	bu2x=Hermitian(mexpi(Lambda2[sx-1][sy]))*mexpi(Lambda2[sx][sy]); 
	bu1y=Hermitian(mexpi(Lambda1[sx][sy-1]))*mexpi(Lambda1[sx][sy]);
	bu2y=Hermitian(mexpi(Lambda2[sx][sy-1]))*mexpi(Lambda2[sx][sy]);
		  
	bu1x/=bu1x.getNorm();
	bu1y/=bu1y.getNorm();
	bu2x/=bu2x.getNorm();
	bu2y/=bu2y.getNorm();
	      
	for (int sz=1; sz<=NUM;sz++)
	  {	    
	    
	    utemp=u2x+u1x*Hermitian(ux[sx][sy][sz])-Hermitian(u2x+u1x*Hermitian(ux[sx][sy][sz]));
	    utemp=utemp+u2y+u1y*Hermitian(uy[sx][sy][sz])-Hermitian(u2y+u1y*Hermitian(uy[sx][sy][sz]));
	    
	    utemp=utemp-bu2x-Hermitian(ux[sx-1][sy][sz])*bu1x+Hermitian(bu2x+Hermitian(ux[sx-1][sy][sz])*bu1x);
	    utemp=utemp-bu2y-Hermitian(uy[sx][sy-1][sz])*bu1y+Hermitian(bu2y+Hermitian(uy[sx][sy-1][sz])*bu1y);
	    
		      //utemp=utemp-(Hermitian(mexpi(Lambda2[sx-1][sy]))*mexpi(Lambda2[sx][sy])+Hermitian(ux[sx-1][sy][sz])*Hermitian(mexpi(Lambda1[sx-1][sy]))*mexpi(Lambda1[sx][sy]))-Hermitian(Hermitian(mexpi(Lambda2[sx-1][sy]))*mexpi(Lambda2[sx][sy])+Hermitian(ux[sx-1][sy][sz])*Hermitian(mexpi(Lambda1[sx-1][sy]))*mexpi(Lambda1[sx][sy]));
		      
		      //utemp=utemp-(Hermitian(mexpi(Lambda2[sx][sy-1]))*mexpi(Lambda2[sx][sy])+Hermitian(uy[sx][sy-1][sz]*Hermitian(mexpi(Lambda1[sx][sy-1]))*mexpi(Lambda1[sx][sy])))-Hermitian(Hermitian(mexpi(Lambda2[sx][sy-1]))*mexpi(Lambda2[sx][sy])+Hermitian(uy[sx][sy-1][sz]*Hermitian(mexpi(Lambda1[sx][sy-1]))*mexpi(Lambda1[sx][sy])));

	    ez[sx][sy][sz].a=utemp.b;
	    ez[sx][sy][sz].b=utemp.c;
	    ez[sx][sy][sz].c=utemp.d;
		      //ez[sx][sy][sz]*=-1;
		  

		      //alpha1x=transportLeft(Lambda1[sx+1][sy]-Lambda1[sx][sy],mexpi(Lambda1[sx][sy]));
		      //alpha2x=transportLeft(Lambda2[sx+1][sy]-Lambda2[sx][sy],mexpi(Lambda2[sx][sy]));
		      //alpha1y=transportLeft(Lambda1[sx][sy+1]-Lambda1[sx][sy],mexpi(Lambda1[sx][sy]));
		      //alpha2y=transportLeft(Lambda2[sx][sy+1]-Lambda2[sx][sy],mexpi(Lambda2[sx][sy]));
		      //ez[sx][sy][sz].print();
		      //cout << "compare to: " << endl;
		      //(icommute(alpha1x,alpha2x)+icommute(alpha1y,alpha2y)).print();
		


		      //utemp=mexpi(Lambda2[sx-1][sy])*mexpi(Lambda2[sx][sy]);
		      //utemp+=Hermitian(u
		      
		      //ez[sx][sy][sz]=icommute(alpha1x,alpha2x)+icommute(alpha1y,alpha2y);
	    resc+=tr2(ez[sx][sy][sz]);
		      //no correction for minus since e_eta in notes is -ez so ok
		      //ez[sx][sy][sz]*=-1;
		      
	  }
      }
  //gaugefix2d(50);
	    
  enforceBCs();

}

void generatefluc()
{
  double sig=XI;
  int l=0;

  cout << "fluctutation ampl " << sig << endl;
  extern double randGauss(double); // generates a gaussian random number with mean 0
  extern void FT1d(double*, int,int);


  for (int sx=0;sx<NUMT;sx++) 
    for (int sy=0;sy<NUMT;sy++)
      {
	if (RANGEN==0)
	  {
	    rho1a[sx*(NUMT)+sy]=randGauss(sig);
	    rho1b[sx*(NUMT)+sy]=randGauss(sig);
	    rho1c[sx*(NUMT)+sy]=randGauss(sig);
	    rho2a[sx*(NUMT)+sy]=randGauss(sig);
	    rho2b[sx*(NUMT)+sy]=randGauss(sig);
	    rho2c[sx*(NUMT)+sy]=randGauss(sig);
	  }
	else
	  {
	    rho1a[sx*(NUMT)+sy]=DoublePrn::getRand()->gaussian_rand_no()*sig;
	    rho1b[sx*(NUMT)+sy]=DoublePrn::getRand()->gaussian_rand_no()*sig;
	    rho1c[sx*(NUMT)+sy]=DoublePrn::getRand()->gaussian_rand_no()*sig;
	    rho2a[sx*(NUMT)+sy]=DoublePrn::getRand()->gaussian_rand_no()*sig;
	    rho2b[sx*(NUMT)+sy]=DoublePrn::getRand()->gaussian_rand_no()*sig;
	    rho2c[sx*(NUMT)+sy]=DoublePrn::getRand()->gaussian_rand_no()*sig;
	  }

	Lambda1[sx+1][sy+1].a=rho1a[sx*(NUMT)+sy];
	Lambda1[sx+1][sy+1].b=rho1b[sx*(NUMT)+sy];
	Lambda1[sx+1][sy+1].c=rho1c[sx*(NUMT)+sy];
	      
	Lambda2[sx+1][sy+1].a=rho2a[sx*(NUMT)+sy];
	Lambda2[sx+1][sy+1].b=rho2b[sx*(NUMT)+sy];
	Lambda2[sx+1][sy+1].c=rho2c[sx*(NUMT)+sy];
      }

  // construct random rapidity profile


  for (int sz=0;sz<NUM;sz++)
    {
      if (RANGEN==0)
	{
	  FT1darray[2*sz]=randGauss(1.0);
	  FT1darray[2*sz+1]=0.0;
	}
      else
	{
	  FT1darray[2*sz]=DoublePrn::getRand()->gaussian_rand_no();
	  FT1darray[2*sz+1]=0.0;
	}
    }

  //check size
  double cr=0;
  for (int sz=0;sz<NUM;sz++)
      cr+=FT1darray[2*sz]*FT1darray[2*sz];
  
  printf("typical fluctuation size: %f\n",sqrt(cr/NUM));


  // can be used to get rid of high momentum modes

  FT1d(FT1darray,NUM,1);

  for (int k=0;k<NUM;k++)
    {
      l=k;
      l=(l<<1);
      if (k>NUM)
	{
	  FT1darray[l]=0.0;
	  l++;
	  FT1darray[l]=0.0;
	}
    }
  
  FT1d(FT1darray,NUM,-1);

  //paste back and generate 2nd function
  for (int sz=0;sz<NUM;sz++)
    {
      f1[sz+1]=FT1darray[2*sz];
      if (RANGEN==0)
	{
	  FT1darray[2*sz]=randGauss(1.0);
	  FT1darray[2*sz+1]=0.0;
	}
      else
	{
	  FT1darray[2*sz]=DoublePrn::getRand()->gaussian_rand_no();
	  FT1darray[2*sz+1]=0.0;
	}
    }

  // get rid of high momentum modes

  FT1d(FT1darray,NUM,1);

  for (int k=0;k<NUM;k++)
    {
      l=k;
      l=(l<<1);
      if (k>NUM)
	{
	  FT1darray[l]=0.0;
	  l++;
	  FT1darray[l]=0.0;
	}
    }
  
  FT1d(FT1darray,NUM,-1);

  //paste back 
  for (int sz=0;sz<NUM;sz++)
    f2[sz+1]=FT1darray[2*sz];



  


  //enforce BC's
  for (int sx=1;sx<=NUMT;sx++) 
    {
      Lambda1[sx][0] = Lambda1[sx][NUMT]; 	Lambda1[sx][NUMT+1] = Lambda1[sx][1];
      Lambda2[sx][0] = Lambda2[sx][NUMT]; 	Lambda2[sx][NUMT+1] = Lambda2[sx][1];
    }
  
  for (int sy=0;sy<=NUMT+1;sy++) 
    {
      Lambda1[0][sy] = Lambda1[NUMT][sy]; 	Lambda1[NUMT+1][sy] = Lambda1[1][sy];
      Lambda2[0][sy] = Lambda2[NUMT][sy]; 	Lambda2[NUMT+1][sy] = Lambda2[1][sy];
    }


  f1[0]=f1[NUM];
  f2[0]=f2[NUM];
  f1[NUM+1]=f1[1];
  f2[NUM+1]=f2[1];


  //this means, pix, piy are ready:

  for (int sx=1;sx<=NUMT;sx++) 
    for (int sy=1;sy<=NUMT;sy++)
      for (int sz=1;sz<=NUM;sz++)
	{
	  pix[sx][sy][sz]=(f1[sz]-f1[sz-1])/A*(Lambda1[sx][sy]-Lambda1[sx][sy-1])+(f2[sz]-f2[sz-1])/A*(Lambda2[sx][sy]-Lambda2[sx][sy-1]);
	  piy[sx][sy][sz]=(f1[sz]-f1[sz-1])/A*(Lambda1[sx][sy]-Lambda1[sx-1][sy])+(f2[sz]-f2[sz-1])/A*(Lambda2[sx][sy]-Lambda2[sx-1][sy]);

	  pix[sx][sy][sz]*=-1.0;
	}


  enforceBCs();


  // test that there is never a divergence left:

  double test=0;

  for (int sx=1;sx<=NUMT;sx++) 
    for (int sy=1;sy<=NUMT;sy++)
      for (int sz=1;sz<=NUM;sz++)
	  test+=tr2(pix[sx][sy][sz]-pix[sx-1][sy][sz]+piy[sx][sy][sz]-piy[sx][sy-1][sz]);

  cout << " Divergence: " << test << endl;

  // now generate ez flucs:

  for (int sx=1;sx<=NUMT;sx++) 
    for (int sy=1;sy<=NUMT;sy++)
      for (int sz=1;sz<=NUM;sz++)
	{
	  ez[sx][sy][sz]+=f1[sz]*(transportLeft(Lambda1[sx-1][sy]-Lambda1[sx-1][sy-1],ux[sx-1][sy][sz])-transportLeft(Lambda1[sx][sy-1]-Lambda1[sx-1][sy-1],uy[sx][sy-1][sz])-(Lambda1[sx][sy]-Lambda1[sx][sy-1])+(Lambda1[sx][sy]-Lambda1[sx-1][sy]));
	  ez[sx][sy][sz]+=f2[sz]*(transportLeft(Lambda2[sx-1][sy]-Lambda2[sx-1][sy-1],ux[sx-1][sy][sz])-transportLeft(Lambda2[sx][sy-1]-Lambda2[sx-1][sy-1],uy[sx][sy-1][sz])-(Lambda2[sx][sy]-Lambda2[sx][sy-1])+(Lambda2[sx][sy]-Lambda2[sx-1][sy]));			  
	}

  enforceBCs();

}

void doMatchingrealistic()
{

  double resa, resb, resc, resd;
  resa=0;
  resb=0;
  resc=0;
  SU2matrix alpha1x,alpha2x,alpha1y,alpha2y;

  SU2matrix temp1,temp2;
	    
  for (int sx=1;sx<=NUMT;sx++) 
    for (int sy=1;sy<=NUMT;sy++)
      {
		  //hold alpha's
	ux[sx][sy][1]=Hermitian(mexpi(Lambda1[sx][sy]))*mexpi(Lambda1[sx+1][sy]);	      
	ux[sx][sy][2]=Hermitian(mexpi(Lambda2[sx][sy]))*mexpi(Lambda2[sx+1][sy]);
	uy[sx][sy][1]=Hermitian(mexpi(Lambda1[sx][sy]))*mexpi(Lambda1[sx][sy+1]);
	uy[sx][sy][2]=Hermitian(mexpi(Lambda2[sx][sy]))*mexpi(Lambda2[sx][sy+1]);
	    
	ux[sx][sy][1]/=ux[sx][sy][1].getNorm(); //reunitarize
	ux[sx][sy][2]/=ux[sx][sy][2].getNorm();
	uy[sx][sy][1]/=uy[sx][sy][1].getNorm();
	uy[sx][sy][2]/=uy[sx][sy][2].getNorm();
      }

  for (int sx=1;sx<=NUMT;sx++) 
    for (int sy=1;sy<=NUMT;sy++)
      {
	alpha1x=Adjoint(ux[sx][sy][1]);
	alpha2x=Adjoint(ux[sx][sy][2]);
	alpha1y=Adjoint(uy[sx][sy][1]);
	alpha2y=Adjoint(uy[sx][sy][2]);
	
	for (int sz=1; sz<=NUM;sz++)
	  {	    
	    //this is Ax
	    temp1=alpha1x*(1-XI*exp(A*(sz-MIDPOINT)))+alpha2x*(1-XI*exp(A*(MIDPOINT-sz)));
		      //resa+=tr2(temp1);
		      
		      //this is Ay
	    temp2=alpha1y*(1-XI*exp(A*(sz-MIDPOINT)))+alpha2y*(1-XI*exp(A*(MIDPOINT-sz)));
		      //resb+=tr2(temp2);


	    u[sx][sy][sz].a = 1;
	    ux[sx][sy][sz]=mexpi(temp1);
	    uy[sx][sy][sz]=mexpi(temp2);
		      
	    ez[sx][sy][sz]=(icommute(alpha1x,alpha2x)+icommute(alpha1y,alpha2y))*(1-XI*2*cosh(A*(sz-MIDPOINT)));
	    resc+=tr2(ez[sx][sy][sz]);
	    pix[sx][sy][sz]=XI*(alpha1x*exp(A*(sz-MIDPOINT))+alpha2x*exp(A*(MIDPOINT-sz)))*(-1.0);
	    piy[sx][sy][sz]=XI*(alpha1y*exp(A*(sz-MIDPOINT))+alpha2y*exp(A*(MIDPOINT-sz)))*(-1.0);
		      //resa+=tr2(icommute(pix[sx][sy][sz],temp1)+icommute(piy[sx][sy][sz],temp2));
		      //resa+=tr2((icommute(alpha1x,alpha2x)+icommute(alpha1y,alpha2y))*XI*2*sinh(A*(sz-MIDPOINT)));   
	    if (sz<NUM)
	      resa+=tr2((icommute(alpha1x,alpha2x)+icommute(alpha1y,alpha2y))*XI*2*(cosh(A*(sz-MIDPOINT))-cosh(A*(sz+1-MIDPOINT))));
	    else 
	      resa+=tr2((icommute(alpha1x,alpha2x)+icommute(alpha1y,alpha2y))*XI*2*(cosh(A*(sz-MIDPOINT))-cosh(A*(1-MIDPOINT))));
		      //no correction for minus since e_eta in notes is -ez so ok
		      //ez[sx][sy][sz]*=-1;
		      
	  }
      }
	    //gaugefix2d(100);

	    //do gauge Transform
	    
  enforceBCs();
  resb=0;
  for (int sx=1;sx<=NUMT;sx++) 
    for (int sy=1;sy<=NUMT;sy++)
      for (int sz=1; sz<=NUM;sz++)
	resb+=tr2((ez[sx][sy][sz-1]-ez[sx][sy][sz])/A);

}




// initializes the variables and then calls enforceBCs()
void setInitialConditions()
{

  double sig;
  extern double randGauss(double); // generates a gaussian random number with mean 0

  if (RANGEN==0)
    {
      sig=TSIG*SQRT2; // standard deviation, normalized as to fit with Yasushi's code
    }
  else
    {
      sig=TSIG*2; // standard deviation
    }
  cout << "checking: " << sig << " vs ." << TSIG << endl;

  //setting MIDPOINT

  //best to have NUM odd

  if (NUM%2!=0)
    MIDPOINT=(NUM+1)/2;
  else
    MIDPOINT=NUM/2;

	//Numerical recipes FFT
	extern void transform(double* pArr,int nx, int ny);

	

    	cout << "==> Initializing variables\n";

        //srand((unsigned int)time(0));
	//srand(19691123);


	//first need to create color sources

	SU2matrix temp1,temp2;
	temp1.a=0;
	temp1.b=0;
	temp1.c=0;
	temp2.a=0;
	temp2.b=0;
	temp2.c=0;


	for (int sx=0;sx<NUMT;sx++) 
	  for (int sy=0;sy<NUMT;sy++)
	    {
	      if (RANGEN==0)
		{
		  rho1a[sx*(NUMT)+sy]=randGauss(sig);
		  rho1b[sx*(NUMT)+sy]=randGauss(sig);
		  rho1c[sx*(NUMT)+sy]=randGauss(sig);
		  rho2a[sx*(NUMT)+sy]=randGauss(sig);
		  rho2b[sx*(NUMT)+sy]=randGauss(sig);
		  rho2c[sx*(NUMT)+sy]=randGauss(sig);
		}
	      else
		{
		  rho1a[sx*(NUMT)+sy]=DoublePrn::getRand()->gaussian_rand_no()*sig;
		  rho1b[sx*(NUMT)+sy]=DoublePrn::getRand()->gaussian_rand_no()*sig;
		  rho1c[sx*(NUMT)+sy]=DoublePrn::getRand()->gaussian_rand_no()*sig;
		  rho2a[sx*(NUMT)+sy]=DoublePrn::getRand()->gaussian_rand_no()*sig;
		  rho2b[sx*(NUMT)+sy]=DoublePrn::getRand()->gaussian_rand_no()*sig;
		  rho2c[sx*(NUMT)+sy]=DoublePrn::getRand()->gaussian_rand_no()*sig;
		}
	      temp1.a+=rho1a[sx*(NUMT)+sy];
	      temp1.b+=rho1b[sx*(NUMT)+sy];
	      temp1.c+=rho1c[sx*(NUMT)+sy];
	      

	      temp2.a+=rho2a[sx*(NUMT)+sy];
	      temp2.b+=rho2b[sx*(NUMT)+sy];
	      temp2.c+=rho2c[sx*(NUMT)+sy];

	    }


	double resa=0;
	double resb=0;
	double resc=0;
	//double resd=0;

	//make it color neutral:

	for (int sx=0;sx<NUMT;sx++) 
	  for (int sy=0;sy<NUMT;sy++)
	    {
	      rho1a[sx*(NUMT)+sy]-=temp1.a/(NUMT)/(NUMT);
	      rho1b[sx*(NUMT)+sy]-=temp1.b/(NUMT)/(NUMT);
	      rho1c[sx*(NUMT)+sy]-=temp1.c/(NUMT)/(NUMT);
	      rho2a[sx*(NUMT)+sy]-=temp2.a/(NUMT)/(NUMT);
	      rho2b[sx*(NUMT)+sy]-=temp2.b/(NUMT)/(NUMT);
	      rho2c[sx*(NUMT)+sy]-=temp2.c/(NUMT)/(NUMT);
	    }

	//check color neutrality

	
	for (int sx=0;sx<NUMT;sx++) 
	  for (int sy=0;sy<NUMT;sy++)
	    {
	      resa+=rho1a[sx*NUMT+sy];
	      resb+=rho1b[sx*NUMT+sy];
	      resc+=rho1c[sx*NUMT+sy];	      
	    }
	if ((resa>1.e-10)&&(resb>1.e-10)&&(resc>1.e-10)) cout << "Violated Color neutrality!" <<endl;
	// cout << "N1:" << resa << " " << resb << " " << resc << endl;

	resa=resb=resc=0;

	for (int sx=0;sx<NUMT;sx++) 
	  for (int sy=0;sy<NUMT;sy++)
	    {
	      resa+=rho2a[sx*NUMT+sy];
	      resb+=rho2b[sx*NUMT+sy];
	      resc+=rho2c[sx*NUMT+sy];	      
	    }
	if ((resa>1.e-10)&&(resb>1.e-10)&&(resc>1.e-10)) cout << "Violated Color neutrality!" <<endl;
	// cout << "N2:" << resa << " " << resb << " " << resc << endl;

	resa=resb=resc=0;
	//calc average rho
	for (int sx=0;sx<NUMT;sx++) 
	  for (int sy=0;sy<NUMT;sy++)
	    {
	      Lambda1[sx+1][sy+1].a=rho1a[sx*(NUMT)+sy];
	      Lambda1[sx+1][sy+1].b=rho1b[sx*(NUMT)+sy];
	      Lambda1[sx+1][sy+1].c=rho1c[sx*(NUMT)+sy];
	      
	      Lambda2[sx+1][sy+1].a=rho2a[sx*(NUMT)+sy];
	      Lambda2[sx+1][sy+1].b=rho2b[sx*(NUMT)+sy];
	      Lambda2[sx+1][sy+1].c=rho2c[sx*(NUMT)+sy];

	      //our solution didn't respect minus in my notes. Correct for it
	      Lambda1[sx+1][sy+1]*=-1;
	      Lambda2[sx+1][sy+1]*=-1;

	      resa+=tr2(Lambda1[sx+1][sy+1]);
	      resb+=tr2(Lambda2[sx+1][sy+1]);
	    }

	// factor 2/3 to accomodate color structure
	cout << "average rho1:" << resa/NUMT/NUMT;
	cout << "  average rho2:" << resb/NUMT/NUMT;
	cout << "  and total av sqrt: " << sqrt((resa+resb)/2/NUMT/NUMT) << endl;

 




	//solve Poisson's eq.
	transform(rho1a,NUMT,NUMT);
	transform(rho1b,NUMT,NUMT);
	transform(rho1c,NUMT,NUMT);

	transform(rho2a,NUMT,NUMT);
	transform(rho2b,NUMT,NUMT);
	transform(rho2c,NUMT,NUMT);

	resa=0;
	resb=0;

	//Paste solution to Lambda
	for (int sx=0;sx<NUMT;sx++) 
	  for (int sy=0;sy<NUMT;sy++)
	    {
	      Lambda1[sx+1][sy+1].a=rho1a[sx*(NUMT)+sy];
	      Lambda1[sx+1][sy+1].b=rho1b[sx*(NUMT)+sy];
	      Lambda1[sx+1][sy+1].c=rho1c[sx*(NUMT)+sy];
	      
	      Lambda2[sx+1][sy+1].a=rho2a[sx*(NUMT)+sy];
	      Lambda2[sx+1][sy+1].b=rho2b[sx*(NUMT)+sy];
	      Lambda2[sx+1][sy+1].c=rho2c[sx*(NUMT)+sy];

	      //our solution didn't respect minus in my notes. Correct for it
	      Lambda1[sx+1][sy+1]*=-1;
	      Lambda2[sx+1][sy+1]*=-1;

	      resa+=tr2(Lambda1[sx+1][sy+1]);
	      resb+=tr2(Lambda2[sx+1][sy+1]);
	    }
	

	//enforce BC's
	for (int sx=1;sx<=NUMT;sx++) 
	  {
	    Lambda1[sx][0] = Lambda1[sx][NUMT]; 	Lambda1[sx][NUMT+1] = Lambda1[sx][1];
	    Lambda2[sx][0] = Lambda2[sx][NUMT]; 	Lambda2[sx][NUMT+1] = Lambda2[sx][1];
	  }

	for (int sy=0;sy<=NUMT+1;sy++) 
	  {
	    Lambda1[0][sy] = Lambda1[NUMT][sy]; 	Lambda1[NUMT+1][sy] = Lambda1[1][sy];
	    Lambda2[0][sy] = Lambda2[NUMT][sy]; 	Lambda2[NUMT+1][sy] = Lambda2[1][sy];
	  }

	cout << "average L1: " << resa/NUMT/NUMT;
	cout << "  average L2: " << resb/NUMT/NUMT << endl;

	resa=0;
	resb=0;


	// IC version I: via direct (continuum) relations
       
	if (MIMICK2D==1)
	  {
	    doMatching2d();

	    generatefluc();
	    otime+=TINIT;
	  }
	
	
	// IC version II: via lattice relations
	if (MIMICK2D==0)
	  {
	    doMatching2d();
	  }
	
	//cout << "checking norm: "  << checknorm/NUMT/NUMT << endl;
	//cout << "checking norm2: "  << checknorm2/NUMT/NUMT << endl;


	cout << "L1: " << resa/NUMT/NUMT;
	cout << "   L2: " << resb/NUMT/NUMT;
	cout << "   Ez: " << resc/NUMT/NUMT/NUM;
	//cout << "   Lin Ez: " << resd/NUMT/NUMT/NUM << endl;

	enforceBCs();
}

SU2matrix phix(int sx, int sy, int sz)
{
SU2matrix temp;
  //temp=A*phix[s];
  temp.a=(Hermitian(ux[sx][sy][sz])-ux[sx][sy][sz]).b;
  temp.b=(Hermitian(ux[sx][sy][sz])-ux[sx][sy][sz]).c;
  temp.c=(Hermitian(ux[sx][sy][sz])-ux[sx][sy][sz]).d;
  return temp;
}

SU2matrix phiy(int sx,int sy,int sz){

  SU2matrix temp;
  //temp=A*phix[s];
  temp.a=(Hermitian(uy[sx][sy][sz])-uy[sx][sy][sz]).b;
  temp.b=(Hermitian(uy[sx][sy][sz])-uy[sx][sy][sz]).c;
  temp.c=(Hermitian(uy[sx][sy][sz])-uy[sx][sy][sz]).d;
  return temp;
}





// computes energy in longitudinal electric fields
double ezEnergyDensity() {
	double res=0;
            // E's and Pi's are defined at half-integer times and the fields 
            // at integer times so in order to most accurately compute the energy density
            // we should average the values of E and Pi at t-eps/2 and t+eps/2
            // to obtain a better approximation to their value at t
	for (int sx=1;sx<=NUMT;sx++)
	  for (int sy=1;sy<=NUMT;sy++)
	    for (int sz=1;sz<=NUM;sz++)
	      {
		res += (tr2(ez[sx][sy][sz]+Ez[sx][sy][sz]))/4;
	      }
	return res/NUM/NUMT/NUMT/TSIG/TSIG/TSIG;
}

// computes energy in transverse electric fields
double etEnergyDensity() {
	double res=0;

            // E's and Pi's are defined at half-integer times and the fields 
            // at integer times so in order to most accurately compute the energy density
            // we should average the values of E and Pi at t-eps/2 and t+eps/2
            // to obtain a better approximation to their value at t

	for (int sx=1;sx<=NUMT;sx++)
	  for (int sy=1;sy<=NUMT;sy++)
	    for (int sz=1;sz<=NUM;sz++)
	      {
		res += (tr2(pix[sx][sy][sz]+Pix[sx][sy][sz]))/4;
		res += (tr2(piy[sx][sy][sz]+Piy[sx][sy][sz]))/4;
	      }
	return res/NUM/NUMT/NUMT/TSIG/TSIG/TSIG;
}

// computes the energy in longitudinal magnetic fields
double bzEnergyDensity() {
	double res=0;
	double tempa;
	SU2LieMatrix dPlaqxy;
	for (int sx=1;sx<=NUMT;sx++)
	  for (int sy=1;sy<=NUMT;sy++)
	    for (int sz=1;sz<=NUM;sz++)
	      {
		// these guys are defined at integer times so no need average here
		dPlaqxy=ux[sx][sy][sz]*uy[sx+1][sy][sz]*Hermitian(ux[sx][sy+1][sz])*Hermitian(uy[sx][sy][sz]);
		tempa=(dPlaqxy.b*dPlaqxy.b+dPlaqxy.c*dPlaqxy.c+dPlaqxy.d*dPlaqxy.d);
		if (tempa<1.e-6)
		  {
		    res+=2*tempa*(1+tempa/4-tempa*tempa/8);
		  }
		else
		  res+=4*(1-dPlaqxy.a);
		//res += tr2(icommute(A*phix(sx,sy,sz),A*phiy(sx,sy,sz)));
		//res+=A*A*tr2(phiy(sx+1,sy,sz)-phiy(sx,sy,sz));
	      }
	return res/NUM/NUMT/NUMT/TSIG/TSIG/TSIG;
}

// computes the energy in transverse magnetic fields
double btEnergyDensity() {
	double res=0;
	SU2LieMatrix dPlaqxz,dPlaqyz;
	double tempb,tempc;

	for (int sx=1;sx<=NUMT;sx++)
	  for (int sy=1;sy<=NUMT;sy++)
	    for (int sz=1;sz<=NUM;sz++)
	      {

		dPlaqxz=ux[sx][sy][sz]*u[sx+1][sy][sz]*Hermitian(ux[sx][sy][sz+1])*Hermitian(u[sx][sy][sz]);
		dPlaqyz=uy[sx][sy][sz]*u[sx][sy+1][sz]*Hermitian(uy[sx][sy][sz+1])*Hermitian(u[sx][sy][sz]);
            // these guys are defined at integer times so no need average here
	    //res += tr2((phix[s]-transportRight(phix[s-1],u[s-1]))/A);
	    //res += tr2((phiy[s]-transportRight(phiy[s-1],u[s-1]))/A);

		tempb=(dPlaqxz.b*dPlaqxz.b+dPlaqxz.c*dPlaqxz.c+dPlaqxz.d*dPlaqxz.d)/A/A;
		tempc=(dPlaqyz.b*dPlaqyz.b+dPlaqyz.c*dPlaqyz.c+dPlaqyz.d*dPlaqyz.d)/A/A;

		if (tempb<1.e-6)
		  {
		    res+=2*tempb*(1+tempb/4-tempb*tempb/8);
		  }
		else
		  res+=4*(1-dPlaqxz.a)/A/A;

		if (tempc<1.e-6)
		  {
		    res+=2*tempc*(1+tempc/4-tempc*tempc/8);
		  }
		else
		  res+=4*(1-dPlaqyz.a)/A/A;
	}
	return res/NUM/NUMT/NUMT/TSIG/TSIG/TSIG;
}

//initial energy density

// update energy loss
// uses midpoint rule to compute integral in def of HL energy
void updateHLenergy(double eps) {
  double tempa=0;
  double tempb=0;
  double tempc=0;
	double eloss=0;
	SU2LieMatrix dPlaqxy,dPlaqxz,dPlaqyz;
	

	for (int sx=1;sx<=NUMT;sx++)
	  for (int sy=1;sy<=NUMT;sy++)
	    for (int sz=1;sz<=NUM;sz++)
	      {	

		dPlaqxy=ux[sx][sy][sz]*uy[sx+1][sy][sz]*Hermitian(ux[sx][sy+1][sz])*Hermitian(uy[sx][sy][sz]);
		dPlaqxz=ux[sx][sy][sz]*u[sx+1][sy][sz]*Hermitian(ux[sx][sy][sz+1])*Hermitian(u[sx][sy][sz]);
		dPlaqyz=uy[sx][sy][sz]*u[sx][sy+1][sz]*Hermitian(uy[sx][sy][sz+1])*Hermitian(u[sx][sy][sz]);

		



		tempa=(dPlaqxy.b*dPlaqxy.b+dPlaqxy.c*dPlaqxy.c+dPlaqxy.d*dPlaqxy.d);
		tempb=(dPlaqxz.b*dPlaqxz.b+dPlaqxz.c*dPlaqxz.c+dPlaqxz.d*dPlaqxz.d)/A/A;
		tempc=(dPlaqyz.b*dPlaqyz.b+dPlaqyz.c*dPlaqyz.c+dPlaqyz.d*dPlaqyz.d)/A/A;
		
		if (tempa<1.e-6)
		  {
		    eloss+=tempa*(1+tempa/4-tempa*tempa/8);
		  }
		else
		  eloss+=2*(1-dPlaqxy.a);
		
		if (otime!=0)
		  {
		    if (tempb<1.e-6)
		      {
			eloss-=tempb*(1+tempb/4-tempb*tempb/8)/otime/otime;
		      }
		    else
		      eloss-=2*(1-dPlaqxz.a)/otime/otime/A/A;
		    
		    if (tempc<1.e-6)
		      {
			eloss-=tempc*(1+tempc/4-tempc*tempc/8)/otime/otime;
		      }
		    else
		      eloss-=2*(1-dPlaqyz.a)/otime/otime/A/A;
		  }

		
		eloss += (tr2(ez[sx][sy][sz]))/2;
		if (otime!=0) eloss -= (tr2(pix[sx][sy][sz])/otime/otime)/2;
		//eloss -= (tr2(Pix[sx][sy][sz])/(otime+2*eps)/(otime+2*eps))/2;
		if (otime!=0) eloss -= (tr2(piy[sx][sy][sz])/otime/otime)/2;
		//eloss -= (tr2(Piy[sx][sy][sz])/(otime+2*eps)/(otime+2*eps))/2;		
	      }
		
	eloss/=NUM*NUMT*NUMT/2*TSIG*TSIG*TSIG;



	//res -= etEnergyDensity()/(otime*otime);
	//res += ezEnergyDensity();
	//res += bzEnergyDensity();
	//res -= btEnergyDensity()/(otime*otime);
	HLenergy += eps*eloss;
	//cout << "in HL" << HLenergy << endl;
}

// a convenience method for getting total energyDensity
double energyDensity() {
	double res = 0;
	if (otime!=0) res += etEnergyDensity()/otime;
	res += ezEnergyDensity()*otime;
	res += bzEnergyDensity()*otime;
	if (otime!=0) res += btEnergyDensity()/otime;
	return res;
}

// computes fluctuations in transverse electric fields
double etflucs() {
  double res=0;

  double meanx=0;
  double meany=0;


  for (int sx=1;sx<=NUMT;sx++)
    for (int sy=1;sy<=NUMT;sy++)
      {
	meanx=0;
	meany=0;
	for (int sz=1;sz<=NUM;sz++)
	  {
	    meanx += tr2(pix[sx][sy][sz])/NUM;
	    meany += tr2(piy[sx][sy][sz])/NUM;
	  }
	for (int sz=1;sz<=NUM;sz++)
	  {
	    res+=(tr2(pix[sx][sy][sz])-meanx)*(tr2(pix[sx][sy][sz])-meanx)+(tr2(piy[sx][sy][sz])-meany)*(tr2(piy[sx][sy][sz])-meany);
	  }
      }

  res/=NUM*NUMT*NUMT;
  
   

  return sqrt(res)/TSIG/TSIG/TSIG;


}


// computes Txx+Tyy/2
double Tperp() {
  double res=0;
  return (ezEnergyDensity()+bzEnergyDensity())*otime;
}

//computes T^\eta\eta
double Tzz() {
  double res=0;
  res+=(btEnergyDensity()+etEnergyDensity())/otime/otime/otime;
  res-=(ezEnergyDensity()+bzEnergyDensity())/otime;
  return res;
}



double checkGaussLaw() {
	double res=0;
	SU2matrix temp;
	temp.a=0; temp.b=0; temp.c=0;
	for (int sx=1;sx<=NUMT;sx++) 
	  for (int sy=1; sy<=NUMT;sy++)
	    for (int sz=1; sz<=NUM; sz++)
	      {
		//temp+=(transportLeft(ez[sx][sy][sz-1],u[sx][sy][sz-1])-ez[sx][sy][sz])/A-(transportLeft(pix[sx-1][sy][sz],ux[sx-1][sy][sz])-pix[sx][sy][sz])-(transportLeft(piy[sx][sy-1][sz],uy[sx][sy-1][sz])-piy[sx][sy][sz]);
		res+=tr2((transportLeft(ez[sx][sy][sz-1],u[sx][sy][sz-1])-ez[sx][sy][sz])/A-(transportLeft(pix[sx-1][sy][sz],ux[sx-1][sy][sz])-pix[sx][sy][sz])-(transportLeft(piy[sx][sy-1][sz],uy[sx][sy-1][sz])-piy[sx][sy][sz]));
	      }
	//res=tr2(temp);
	return sqrt(res/NUM/NUMT/NUMT);
}

double phiRMS() {
	double res=0;
	for (int sx=1;sx<=NUMT;sx++) 
	  for (int sy=1;sy<=NUMT;sy++)
	    for (int sz=1; sz<=NUM;sz++)  
	      { res += tr2(phix(sx,sy,sz))+tr2(phiy(sx,sy,sz)); }
	return sqrt(2*res/NUMT/NUMT/NUM);
}

double piRMS() {
	double res=0;
	for (int sx=1;sx<=NUMT;sx++) 
	  for (int sy=1;sy<=NUMT;sy++)
	    for (int sz=1; sz<=NUM;sz++) 
	      { res += tr2(pix[sx][sy][sz])+tr2(piy[sx][sy][sz]); }
	return sqrt(2*res/NUMT/NUMT/NUM/A/A);
}

double ezRMS() {
	double res=0;
	for (int sx=1;sx<=NUMT;sx++) 
	  for (int sy=1;sy<=NUMT;sy++)
	    for (int sz=1; sz<=NUM;sz++) 
	      { res += tr2(ez[sx][sy][sz]); }
	return sqrt(2*res/NUMT/NUMT/NUM/A/A/A/A);
}

double j0RMS() {
	double res = 0;
	SU2matrix temp;
	for (int sx=1;sx<=NUMT;sx++) 
	  for (int sy=1;sy<=NUMT;sy++)
	    for (int sz=1; sz<=NUM;sz++) { 
	}
	res/=NUMT*A*A;
	res/=NUMT*A*A;
	res/=NUM*A*A;
	return sqrt(2*res);
}

double jzRMS() {
        // remember:  j[NUMT+2][3]
	double res=0;
	for (int sx=1;sx<=NUMT;sx++) 
	  for (int sy=1;sy<=NUMT;sy++)
	    for (int sz=1; sz<=NUM;sz++) { 
	      res += tr2(j[sx][sy][sz][2]); }
	res/=NUMT*A*A;
	res/=NUMT*A*A;
	res/=NUM*A*A;
	return sqrt(2*res);
}

double jtRMS() {
        // remember:  j[NUMT+2][3]
	double res=0;
	for (int sx=1;sx<=NUMT;sx++) 
	  for (int sy=1;sy<=NUMT;sy++)
	    for (int sz=1; sz<=NUM;sz++) { 
	      res += tr2(j[sx][sy][sz][0])+tr2(j[sx][sy][sz][1]); }
	res/=NUMT*A*A;
	res/=NUMT*A*A;
	res/=NUM*A*A;
	return sqrt(2*res);
}

// computes phi-based C
double C(int sx, int sy, int sz) {
	// note that this is not well-defined if phi's are vanishing
  return sqrt(tr2(icommute(phix(sx,sy,sz),phiy(sx,sy,sz))))/(tr2(phix(sx,sy,sz))+tr2(phiy(sx,sy,sz)));
}

// computes phi-based Cbar
double Cbar() {
	double res=0;
	for (int sx=1;sx<=NUMT;sx++) 
	  for (int sy=1;sy<=NUMT;sy++)
	    for (int sz=1; sz<=NUM;sz++) { res += C(sx,sy,sz); }
	return res/NUMT/NUMT/NUM;
}

// computes j-based C
double Cj(int sx, int sy, int sz) {
	// note j[NUMT+2][3]
	// note that this is not well-defined if j's are vanishing
  return sqrt(tr2(icommute(j[sx][sy][sz][0],j[sx][sy][sz][1])))/(tr2(j[sx][sy][sz][0])+tr2(j[sx][sy][sz][1]));
}

// computes j-based Cbar
double Cjbar() {
	double res=0;
	for (int sx=1;sx<=NUMT;sx++) 
	  for (int sy=1;sy<=NUMT;sy++)
	    for (int sz=1; sz<=NUM;sz++) { 
	      res += Cj(sx,sy,sz); }
	return res/NUMT/NUMT/NUM;
}

// computes xiA, xiPhi, xiAJ, and xiJ
void computeCorrelationLengths() {
	
	// first initialize uxiphi's and uxij's to their values at xi=0 
  for (int sx=0;sx<=NUMT+1;sx++) 
    for (int sy=0;sy<=NUMT+1;sy++)
      for (int sz=0; sz<=NUM+1;sz++) { 
	uxiphix[sx][sy][sz]=phix(sx,sy,sz); uxiphiy[sx][sy][sz]=phiy(sx,sy,sz); 
	uxijx[sx][sy][sz]=j[sx][sy][sz][0]; uxijy[sx][sy][sz]=j[sx][sy][sz][1]; 
      }
	// compute the norm needed for chiPhixi and chiJxi
	double norm=0,normJ=0; 
	for (int sx=1;sx<=NUMT;sx++) 
	  for (int sy=1;sy<=NUMT;sy++)
	    for (int sz=1; sz<=NUM;sz++) { 
	      norm += tr2(phix(sx,sy,sz))+tr2(phiy(sx,sy,sz)); 
	      normJ += tr2(j[sx][sy][sz][0])+tr2(j[sx][sy][sz][1]);
	    } 
	norm/=NUMT*NUMT*NUM; normJ/=NUMT*NUMT*NUM;

	// now loop over x until all xi's are computed
  	bool cA=false,cP=false,cAJ=false,cJ=false;
	int sx=3; 
	int sy=3;
	for (int x=1;x<=NUMT;x++) { 	
	  // update the parallel transported fields
		for (int sz=0;sz<NUM;sz++) {
		  uxiphix[sx][sy][sz]=transportRight(uxiphix[sx][sy][sz],u[sx][sy][(sz+x-1)%NUM]);
		  uxiphiy[sx][sy][sz]=transportRight(uxiphiy[sx][sy][sz],u[sx][sy][(sz+x-1)%NUM]);
		  uxijx[sx][sy][sz]=transportRight(uxijx[sx][sy][sz],u[sx][sy][(sz+x-1)%NUM]);
		  uxijy[sx][sy][sz]=transportRight(uxijy[sx][sy][sz],u[sx][sy][(sz+x-1)%NUM]);
		}

		//this is too complicated for now: skip it!
		break;
		/*
	  	// now compute the integral over s
	  	double chiAxi=0,chiPhixi=0,chiAJxi=0,chiJxi=0;
		for (int s=0;s<NUM;s++) {
			double temp = 0;
			if (!cA) {
				temp  = tr2(icommute(phix[(s+x)%NUM],uxiphix[s]));
				temp += tr2(icommute(phix[(s+x)%NUM],uxiphiy[s]));
				temp += tr2(icommute(phiy[(s+x)%NUM],uxiphix[s]));
				temp += tr2(icommute(phiy[(s+x)%NUM],uxiphiy[s]));
				temp /= (tr2(phix[(s+x)%NUM])+tr2(phiy[(s+x)%NUM]));
				temp /= (tr2(phix[s])+tr2(phiy[s]));
				chiAxi += temp*0.75; // d_A/(2C_A)=3/4 for SU(2)
			}
			if (!cP) {
				chiPhixi += (tr(phix[(s+x)%NUM],uxiphix[s])+tr(phiy[(s+x)%NUM],uxiphiy[s]));
			}
			if (!cAJ) {
				temp  = tr2(icommute(j[(s+x)%NUM][0],uxijx[s]));
				temp += tr2(icommute(j[(s+x)%NUM][0],uxijy[s]));
				temp += tr2(icommute(j[(s+x)%NUM][1],uxijx[s]));
				temp += tr2(icommute(j[(s+x)%NUM][1],uxijy[s]));
				temp /= (tr2(j[(s+x)%NUM][0])+tr2(j[(s+x)%NUM][1]));
				temp /= (tr2(j[s][0])+tr2(j[s][1]));
				chiAJxi += temp*0.75; // d_A/(2C_A)=3/4 for SU(2)
			}
			if (!cJ) {
				chiJxi += (tr(j[(s+x)%NUM][0],uxijx[s])+tr(j[(s+x)%NUM][1],uxijy[s]));
			}
		}
		if (!cA) {
			// normalize
			chiAxi/=NUM;
			// and then check size
			if (chiAxi>=0.5) { xiA=x*A; cA=true; }
		}
		if (!cP) { 
			// normalize
			chiPhixi/=norm; chiPhixi/=NUM; 
			// and then check size
			if (fabs(chiPhixi)<0.5) { xiPhi=x*A; cP=true; }
		}
		if (!cAJ) {
			// normalize
			chiAJxi/=NUM;
			// and then check size
			if (chiAJxi>=0.5) { xiAJ=x*A; cAJ=true; }
		}
		if (!cJ) { 
			// normalize
			chiJxi/=normJ; chiJxi/=NUM; 
			// and then check size
			if (fabs(chiJxi)<0.5) { xiJ=x*A; cJ=true; }
		}		
		// if everything has been calculated break the loop and return
		if (cA&&cP&&cAJ&&cJ) break;
		*/
	}
}




//main update routine
inline void doInc(double eps,double epsn) {
  //SU2matrix c,d;
  SU2matrix temp;
  SU2LieMatrix Ua,Ub,Uc,Ud,Utemp;
  //loadjs();


  if (otime==0)
    for (int sx=0;sx<=NUMT+1;sx++) 
      for (int sy=0;sy<=NUMT+1;sy++)
	for (int sz=0; sz<=NUM+1;sz++) { 
	  u[sx][sy][sz] = mexpi(eps*ez[sx][sy][sz]*(otime+eps/2)*A)*u[sx][sy][sz]; 
	  ux[sx][sy][sz] = mexpi(eps*pix[sx][sy][sz]*(-1.0)/(otime+eps/2))*ux[sx][sy][sz];
	  uy[sx][sy][sz] = mexpi(eps*piy[sx][sy][sz]*(-1.0)/(otime+eps/2))*uy[sx][sy][sz];
	}


	for (int sx=1;sx<=NUMT;sx++) 
	  for (int sy=1;sy<=NUMT;sy++)
	    for (int sz=1; sz<=NUM;sz++) { 


	      //update momenta
	      Ua=ux[sx][sy][sz]*uy[sx+1][sy][sz]*Hermitian(ux[sx][sy+1][sz])*Hermitian(uy[sx][sy][sz]);
	      Ub=ux[sx][sy][sz]*Hermitian(uy[sx+1][sy-1][sz])*Hermitian(ux[sx][sy-1][sz])*uy[sx][sy-1][sz];
	      Uc=ux[sx][sy][sz]*u[sx+1][sy][sz]*Hermitian(ux[sx][sy][sz+1])*Hermitian(u[sx][sy][sz]);
	      Ud=ux[sx][sy][sz]*Hermitian(u[sx+1][sy][sz-1])*Hermitian(ux[sx][sy][sz-1])*u[sx][sy][sz-1];


	      temp=Adjoint(Ua+Ub)*(otime+eps)+Adjoint(Uc+Ud)/(otime+eps)/A/A;

	      Pix[sx][sy][sz] = pix[sx][sy][sz] + 2*eps*temp;

	      Ua=Hermitian(Ua);
	      Utemp=Uc;
	      Ub=uy[sx][sy][sz]*Hermitian(ux[sx-1][sy+1][sz])*Hermitian(uy[sx-1][sy][sz])*ux[sx-1][sy][sz];
	      Uc=uy[sx][sy][sz]*u[sx][sy+1][sz]*Hermitian(uy[sx][sy][sz+1])*Hermitian(u[sx][sy][sz]);
	      Ud=uy[sx][sy][sz]*Hermitian(u[sx][sy+1][sz-1])*Hermitian(uy[sx][sy][sz-1])*u[sx][sy][sz-1];
	      
	      temp=Adjoint(Ua+Ub)*(otime+eps)+Adjoint(Uc+Ud)/(otime+eps)/A/A;

	      Piy[sx][sy][sz] = piy[sx][sy][sz] + 2*eps*temp;


	      Ua=Hermitian(Utemp);
	      Ub=u[sx][sy][sz]*Hermitian(ux[sx-1][sy][sz+1])*Hermitian(u[sx-1][sy][sz])*ux[sx-1][sy][sz];
	      Uc=Hermitian(Uc);
	      Ud=u[sx][sy][sz]*Hermitian(uy[sx][sy-1][sz+1])*Hermitian(u[sx][sy-1][sz])*uy[sx][sy-1][sz];

	      temp=Adjoint(Ua+Ub+Uc+Ud)/A;
		
	      Ez[sx][sy][sz] = ez[sx][sy][sz] - 2*eps/(otime+eps)*temp;



	      // then use the values just obtained to update fields
		

	      
	      U[sx][sy][sz] = mexpi((eps+epsn)*Ez[sx][sy][sz]*(otime+eps)*A)*u[sx][sy][sz]; 
	      Ux[sx][sy][sz] = mexpi((eps+epsn)*Pix[sx][sy][sz]*(-1.0)/(otime+eps))*ux[sx][sy][sz];
	      Uy[sx][sy][sz] = mexpi((eps+epsn)*Piy[sx][sy][sz]*(-1.0)/(otime+eps))*uy[sx][sy][sz];
	      
	    }
}


inline void firstStep(double eps) {
  //SU2matrix c,d;
  SU2matrix temp;
  SU2LieMatrix Ua,Ub,Uc,Ud,Utemp;
  //loadjs();


    for (int sx=0;sx<=NUMT+1;sx++) 
      for (int sy=0;sy<=NUMT+1;sy++)
	for (int sz=0; sz<=NUM+1;sz++) { 
	  u[sx][sy][sz] = mexpi(eps*ez[sx][sy][sz]*(eps/2)*A)*u[sx][sy][sz]; 
	  ux[sx][sy][sz] = ux[sx][sy][sz];
	  uy[sx][sy][sz] = uy[sx][sy][sz];
	}

    for (int sx=1;sx<=NUMT;sx++) 
      for (int sy=1;sy<=NUMT;sy++)
	for (int sz=1; sz<=NUM;sz++) { 
	  
	  Ua=ux[sx][sy][sz]*uy[sx+1][sy][sz]*Hermitian(ux[sx][sy+1][sz])*Hermitian(uy[sx][sy][sz]);
	  Ub=ux[sx][sy][sz]*Hermitian(uy[sx+1][sy-1][sz])*Hermitian(ux[sx][sy-1][sz])*uy[sx][sy-1][sz];
	  Uc=ux[sx][sy][sz]*u[sx+1][sy][sz]*Hermitian(ux[sx][sy][sz+1])*Hermitian(u[sx][sy][sz]);
	  Ud=ux[sx][sy][sz]*Hermitian(u[sx+1][sy][sz-1])*Hermitian(ux[sx][sy][sz-1])*u[sx][sy][sz-1];
	  

	  
	  temp=Adjoint(Ua+Ub)*(otime+eps)+Adjoint(Uc+Ud)/(otime+eps)/A/A;

	  Pix[sx][sy][sz] = pix[sx][sy][sz] + 2*eps*temp;
	  
	  Ua=Hermitian(Ua);
	  Utemp=Uc;
	  Ub=uy[sx][sy][sz]*Hermitian(ux[sx-1][sy+1][sz])*Hermitian(uy[sx-1][sy][sz])*ux[sx-1][sy][sz];
	  Uc=uy[sx][sy][sz]*u[sx][sy+1][sz]*Hermitian(uy[sx][sy][sz+1])*Hermitian(u[sx][sy][sz]);
	  Ud=uy[sx][sy][sz]*Hermitian(u[sx][sy+1][sz-1])*Hermitian(uy[sx][sy][sz-1])*u[sx][sy][sz-1];
	  
	  temp=Adjoint(Ua+Ub)*(otime+eps)+Adjoint(Uc+Ud)/(otime+eps)/A/A;
	  
	  Piy[sx][sy][sz] = piy[sx][sy][sz] + 2*eps*temp;


	  Ua=Hermitian(Utemp);
	  Ub=u[sx][sy][sz]*Hermitian(ux[sx-1][sy][sz+1])*Hermitian(u[sx-1][sy][sz])*ux[sx-1][sy][sz];
	  Uc=Hermitian(Uc);
	  Ud=u[sx][sy][sz]*Hermitian(uy[sx][sy-1][sz+1])*Hermitian(u[sx][sy-1][sz])*uy[sx][sy-1][sz];
	  
	  temp=Adjoint(Ua+Ub+Uc+Ud)/A;
		
	  Ez[sx][sy][sz] = ez[sx][sy][sz] - 2*eps/(otime+eps)*temp;



	  U[sx][sy][sz] = mexpi(eps*Ez[sx][sy][sz]*(3*eps/4)*A)*u[sx][sy][sz]; 
	  Ux[sx][sy][sz] = mexpi(eps*Pix[sx][sy][sz]*(-1.0)/(3*eps/4))*ux[sx][sy][sz];
	  Uy[sx][sy][sz] = mexpi(eps*Piy[sx][sy][sz]*(-1.0)/(3*eps/4))*uy[sx][sy][sz];
	      
	}

    //copyDown();
    //enforceBCs();

}


// outputMeasurements and snapshot are defined here
#include "outputroutines.cpp"


//main driver routine
void Evolve() {
  
  //setting step sizes to maximum step size
  double eps = EPS;
  double epsn = eps;


	//if 3d simulation, use adaptive step size, see below, too
	if (MIMICK2D==1)
	  {

	    //eps = energyDensity()/60;
	    //epsn = eps;

	    //this assumes we start at a finite otime already!
	    
	    eps = EPS*otime/(MU2+otime);
	    epsn = eps;
	    cout << "Starting with " << eps << endl;
	  }

	//getting started
	//loadjs();
	//savejs();

	for (int i=1;i<=STEPS;i++) {
		// evolve fields eps forward in time storing updated fields in captial vars

	  
	  

	  // update energy loss
                
	  if (otime!=0) updateHLenergy(eps);

	  
	  doInc(eps,epsn); 

		// measurements and data dump
		//cout << "just for fun: " << HLenergy << endl; 
		if (i==2) 
		  {
		    outputMeasurements(otime);
		  }
		if ( (i>1 && (i-1)%UPDATE==0)) {		  
		  outputMeasurements(otime);
		}
		if ((i-1)%SNAPUPDATE==0) {
		  
		  if (MIMICK2D==0)
		    gaugefix2d(200);
		  else 
		    gaugefix3d(400,otime);
		  
		  snapshot(otime); 
		}

		
		
		//copy fields from capital vars to lowercase vars
		copyDown();
		
		
		// increment time
		otime += 2*eps;

		//again (second part)
		updateHLenergy(eps);


		//if ( (i>1 && (i-1)%UPDATE==0)) {		  
		if (MIMICK2D==1)
		    {
		      eps =epsn;
		      //works very well, but time consuming to compute
		      //epsn = energyDensity()/60.0;

		      
		      //better? reaches EPS at late times
		      epsn = EPS*otime/(MU2+otime);
		    }
		//}

	}
}

void printDivider() {
	int dwidth = 13;
	for (int i=0;i<8*dwidth;i++) cout << "-"; cout << endl;
	return;
}

int main() {
	
	extern void readParameters(char*);

	// open files to put the data in first	
	phirms_out.open("data/phirms.dat", ios::out);
	ED_out.open("data/ED.dat", ios::out);
	jrms_out.open("data/jrms.dat", ios::out);
	cbar_out.open("data/cbar.dat", ios::out);
        energy_out.open("data/energy.dat", ios::out);
	gausslaw_out.open("data/gausslaw.dat", ios::out);
	clengths_out.open("data/clengths.dat", ios::out);
	
	printDivider();
	
	readParameters("params.txt");
	

	printDivider();

	allocateMemory();
		
	//initializeVelocitiesAndCoefficients();


	if (RANGEN==0)
	  {
	    if (RANDOM==0)
	      srand((unsigned int)time(0));
	    else 
	      srand(19771022);
	  }
	else 
	  {
	    if (RANDOM==0)
	      DoublePrn::setRand((unsigned int)time(0));
	    else 
	      DoublePrn::setRand(45030973);
	  }
	
	for (int i=1; i<= NTRAJ; i++)
	  {
	    
	    setInitialConditions();
	    
	    //snapshot(otime);
	
	    printDivider();

	    cout << "Spatial Step Size (A): " << A << endl;
	    
	    cout << "Adaptive stepsize algorithm" << endl;

	    cout << "Maximal Temporal Step Size (EPS): " << EPS << endl;
	    
	    cout << "Longitudinal Box Size (A*" << NUM << "): " << A*NUM << endl;
	    
	    cout << "Transverse Box Size (A*" << NUMT << "): " << A*NUMT << endl;

	    printDivider();



	    double phirms = btEnergyDensity();
	    cout << "Initial Bt ED: " << phirms << endl;
	    
	    double phizrms = bzEnergyDensity();
	    cout << "Initial Bz ED:" << phizrms << endl;
	    
	    double etrms = etEnergyDensity()*4;
	    cout << "Initial et ED: " << etrms << endl;
	    
	    double ezrms = ezEnergyDensity()*4;
	    cout << "Initial ez ED: " << ezrms << endl;
	
	    //InitialEnergy = 4*energyDensity();
	    InitialEnergy = 2*energyDensity();
	    cout << "Initial Energy Density: " << InitialEnergy << endl;

	    //Initial GL
	    double gauss=checkGaussLaw();
	    cout << "Initial Gauss Law: " << gauss << endl;
	  

	    printDivider();
	
	    int dwidth = 13;
	    cout.width(dwidth); cout << "Time";
	    cout.width(dwidth); cout << "<Bz>";
	    cout.width(dwidth); cout << "<Bt>";
	    cout.width(dwidth); cout << "<Ez>";
	    cout.width(dwidth); cout << "<Et>";	
	    cout.width(dwidth); cout << "Energy";
	    cout.width(dwidth); cout << "GaussCheck";
	    cout.width(dwidth); cout << "EnergyViol";
	    cout << endl;
	    
	    printDivider();
	    
	    Evolve();
	    
	    printDivider();
	    cout << "Done.\n";
	    printDivider();
	    
	    // finally close files
	    phirms_out.close();
	    ED_out.close();
	    jrms_out.close();
	    cbar_out.close();
	    energy_out.close();
	    gausslaw_out.close();
	    clengths_out.close();
	  }
	return 0;
}
