using namespace std;

/* 
 * SU2 Lie Group Matrix Class
 *
 * All elements in the Lie Group of SU(2) can be expressed as
 * 
 *   U = a I + i b s_1 + i c s_2 + i d s_3
 *
 * Where I is the identity matrix and s_i are the Pauli matrices
 *
*/

class SU2LieMatrix {

public:    

    double a,b,c,d;
    
    SU2LieMatrix(double a, double b, double c, double d)
    {
        this->a = a; 
	this->b = b; 
	this->c = c;
	this->d = d;
    }

    SU2LieMatrix()
    { 
	this->a = 0; 
	this->b = 0; 
	this->c = 0;
	this->d = 0;
    }

    void normalize()
    {
	double temp = sqrt(this->a*this->a + this->b*this->b + this->c*this->c + this->d*this->d);
	this->a /= temp;
	this->b /= temp;
	this->c /= temp;
	this->d /= temp;
    }

    double getNorm()
    {
	return sqrt(this->a*this->a + this->b*this->b + this->c*this->c + this->d*this->d);
    }

    SU2LieMatrix operator+(const SU2LieMatrix &m)
    {
        SU2LieMatrix temp;
        temp.a = this->a + m.a; 
	temp.b = this->b + m.b; 
	temp.c = this->c + m.c;
	temp.d = this->d + m.d;
        return temp;
    }

    SU2LieMatrix operator-(const SU2LieMatrix &m)
    {
        SU2LieMatrix temp;
        temp.a = this->a - m.a; 
	temp.b = this->b - m.b; 
	temp.c = this->c - m.c;
	temp.d = this->d - m.d;
        return temp;
    }

    SU2LieMatrix operator/(const int &i)
    {
        SU2LieMatrix temp;
        temp.a = this->a/i;
        temp.b = this->b/i;
        temp.c = this->c/i;
        temp.d = this->d/i;
        return temp;
    }

    SU2LieMatrix operator/(const double &d)
    {
        SU2LieMatrix temp;
        temp.a = this->a/d;
        temp.b = this->b/d;
        temp.c = this->c/d;
        temp.d = this->d/d;
        return temp;
    }

    SU2LieMatrix operator/=(const double &d)
    {
        this->a /= d;
        this->b /= d;
        this->c /= d;
	this->d /= d;
        return *this;
    }

    SU2LieMatrix operator*(const SU2LieMatrix &m)
    {
        SU2LieMatrix temp;
	temp.a = this->a*m.a - this->b*m.b - this->c*m.c - this->d*m.d;
	temp.b = this->a*m.b + this->b*m.a - this->c*m.d + this->d*m.c;
	temp.c = this->a*m.c + this->b*m.d + this->c*m.a - this->d*m.b;
	temp.d = this->a*m.d - this->b*m.c + this->c*m.b + this->d*m.a;
	return temp;
    }


    /* used for testing 
    gaugeMat toMatrix() {

    	gaugeMat id,s1,s2,s3,temp;
	dcomp i(0,1);

	id = 1,0,0,1; 
	s1 = 0,i,i,0;
	s2 = 0,1,-1,0;
	s3 = i,0,0,-i;

	temp  = id*this->a;
	temp += s1*this->b;
	temp += s2*this->c;
	temp += s3*this->d;

	return temp;
    }*/

    void print()
    {
	cout.precision(20);
        cout << "(" << this->a << "," << this->b << "," << this->c << "," << this->d;
	cout << ") : norm-1=" << getNorm()-1.0 << endl;
    }
    
private:
   

};

/* 
 * SU2 Algebra Matrix Class
 *
 * All elements in the SU(2) Algebra can be expressed as
 * 
 *   U = a t_1 + b t_2 + c t_3
 *
 * Where t_i are the SU(2) generators which are given by 
 * 1/2 of the Pauli matrices used above, i.e. t_i = 1/2 s_i.
 *
*/

class SU2matrix {

public:    

    double a,b,c;
    
    SU2matrix(double a, double b, double c)
    {
        this->a = a; 
	this->b = b; 
	this->c = c;
    }

    SU2matrix()
    { 
	this->a = 0; 
	this->b = 0; 
	this->c = 0;
    }

    SU2matrix operator*(const double &d)
    {
        SU2matrix temp;
        temp.a = d*this->a;
        temp.b = d*this->b;
        temp.c = d*this->c;
        return temp;
    }

    SU2matrix operator*=(const double &d)
    {
        this->a *= d;
        this->b *= d;
        this->c *= d;
        return *this;
    }

    SU2matrix operator/(const double &d)
    {
        SU2matrix temp;
        temp.a = this->a/d;
        temp.b = this->b/d;
        temp.c = this->c/d;
        return temp;
    }

    SU2matrix operator/=(const double &d)
    {
        this->a /= d;
        this->b /= d;
        this->c /= d;
        return *this;
    }

    SU2matrix operator*(const int &i)
    {
        SU2matrix temp;
        temp.a = i*this->a;
        temp.b = i*this->b;
        temp.c = i*this->c;
        return temp;
    }

    SU2matrix operator*=(const int &i)
    {
        this->a *= i;
        this->b *= i;
        this->c *= i;
        return *this;
    }

    SU2matrix operator/(const int &i)
    {
        SU2matrix temp;
        temp.a = this->a/i;
        temp.b = this->b/i;
        temp.c = this->c/i;
        return temp;
    }

    SU2matrix operator/=(const int &i)
    {
        this->a /= i;
        this->b /= i;
        this->c /= i;
        return *this;
    }

    SU2matrix operator+(const SU2matrix &m)
    {
        SU2matrix temp;
        temp.a = this->a + m.a; 
	temp.b = this->b + m.b; 
	temp.c = this->c + m.c;
        return temp;
    }

    SU2matrix operator+=(const SU2matrix &m)
    {
        this->a += m.a;
        this->b += m.b;
        this->c += m.c;
        return *this;
    }

    SU2matrix operator-(const SU2matrix &m)
    {
        SU2matrix temp;
        temp.a = this->a - m.a; 
	temp.b = this->b - m.b; 
	temp.c = this->c - m.c;
        return temp;
    }

    SU2matrix operator-=(const SU2matrix &m)
    {
        this->a -= m.a;
        this->b -= m.b;
        this->c -= m.c;
        return *this;
    }

    /* used for testing 
    gaugeMat toMatrix() {

    	gaugeMat t1,t2,t3,temp;
	dcomp i(0,1);

	t1 = 0,0.5,0.5,0;
	t2 = 0,-0.5*i,0.5*i,0;
	t3 = 0.5,0,0,-0.5;

	temp  = t1*this->a;
	temp += t2*this->b;
	temp += t3*this->c;
	
	return temp;
    }*/

    void print()
    {
        cout << "(" << a << "," << b << "," << c << ")\n";
    }
    
private:
    
};

SU2matrix operator*(const double d,const SU2matrix m)
{
	SU2matrix temp;
        temp.a = d*m.a;
        temp.b = d*m.b;
        temp.c = d*m.c;
        return temp;
}

SU2matrix operator*(const int i,const SU2matrix m)
{
	SU2matrix temp;
        temp.a = i*m.a;
        temp.b = i*m.b;
        temp.c = i*m.c;
        return temp;
}



//computes Hermitian conjugate of Lie Matrix
SU2LieMatrix Hermitian(SU2LieMatrix m)
{
  SU2LieMatrix temp;
  temp.a=m.a;
  temp.b=-m.b;
  temp.c=-m.c;
  temp.d=-m.d;
  return temp;
}

//computes SU2 Matrix from Lie Matrix
SU2matrix Adjoint(SU2LieMatrix m)
{
  SU2matrix temp;
  temp.a=2*m.b;
  temp.b=2*m.c;
  temp.c=2*m.d;
  return temp;
}



//
// transport left computes Udagger.M.U 
//
SU2matrix transportLeft(SU2matrix m,SU2LieMatrix u) {
	SU2matrix temp;
	temp.a = (2*u.a*u.a-1)*m.a + 2*u.b*(u.b*m.a + u.c*m.b + u.d*m.c) - 2*u.a*(m.b*u.d-m.c*u.c);
	temp.b = (2*u.a*u.a-1)*m.b + 2*u.c*(u.b*m.a + u.c*m.b + u.d*m.c) - 2*u.a*(m.c*u.b-m.a*u.d);
	temp.c = (2*u.a*u.a-1)*m.c + 2*u.d*(u.b*m.a + u.c*m.b + u.d*m.c) - 2*u.a*(m.a*u.c-m.b*u.b);
	return temp;
}

//
// transport right computes U.M.Udagger
//
SU2matrix transportRight(SU2matrix m,SU2LieMatrix u) {
	SU2matrix temp;
	temp.a = (2*u.a*u.a-1)*m.a + 2*u.b*(u.b*m.a + u.c*m.b + u.d*m.c) + 2*u.a*(m.b*u.d-m.c*u.c);
	temp.b = (2*u.a*u.a-1)*m.b + 2*u.c*(u.b*m.a + u.c*m.b + u.d*m.c) + 2*u.a*(m.c*u.b-m.a*u.d);
	temp.c = (2*u.a*u.a-1)*m.c + 2*u.d*(u.b*m.a + u.c*m.b + u.d*m.c) + 2*u.a*(m.a*u.c-m.b*u.b);
	return temp;
}


double tr2(SU2matrix m) {
    	return (m.a*m.a + m.b*m.b + m.c*m.c)/2;
}

double tr(SU2matrix m,SU2matrix n) {
    	return (m.a*n.a + m.b*n.b + m.c*n.c)/2;
}

// computes i times commuator of two SU2 matrices
SU2matrix icommute(SU2matrix u,SU2matrix v) {
	SU2matrix temp;
	temp.a = u.c*v.b - u.b*v.c;
	temp.b = u.a*v.c - u.c*v.a;
	temp.c = u.b*v.a - u.a*v.b;
	return temp;
}

// computes i[u,i[u,v]]
SU2matrix cuuv(SU2matrix u,SU2matrix v) {
	SU2matrix temp;
	temp.a  = u.a*(u.b*v.b+u.c*v.c);
	temp.a -= v.a*(u.b*u.b+u.c*u.c);
	temp.b  = u.b*(u.a*v.a+u.c*v.c);
	temp.b -= v.b*(u.a*u.a+u.c*u.c);
	temp.c  = u.c*(u.a*v.a+u.b*v.b);
	temp.c -= v.c*(u.a*u.a+u.b*u.b);
	return temp;
}

// computes a Lie element from exp(i SU2matrix) via analytic formula
SU2LieMatrix mexpi(SU2matrix m) {
	SU2LieMatrix lm;
	double x = sqrt(m.a*m.a + m.b*m.b + m.c*m.c);	
	if (x==0) { lm.a=1; lm.b=0; lm.c=0; lm.d=0; return lm; }
	lm.a = cos(x/2);
	double s = sin(x/2); 
	//double s = sqrt(1-lm.a*lm.a); 
	lm.b = m.a*s/x; lm.c = m.b*s/x; lm.d = m.c*s/x;
	return lm;
}
